"""Cardano network types."""

from __future__ import annotations

from enum import Enum
from typing import Type

from pycardano.serialization import CBORSerializable, limit_primitive_type

__all__ = ["Network"]


class Network(CBORSerializable, Enum):
    """
    Network ID
    """

    TESTNET = 0
    MAINNET = 1

    def to_primitive(self) -> int:
        return self.value

    @classmethod
    @limit_primitive_type(int)
    def from_primitive(cls: Type[Network], value: int) -> Network:
        return cls(value)



"""A collection of utility functions."""

from __future__ import annotations

import math
import sys
from typing import Dict, List, Optional, Tuple, Union

from nacl.encoding import RawEncoder
from nacl.hash import blake2b

from pycardano.backend.base import ChainContext
from pycardano.cbor import cbor2
from pycardano.hash import SCRIPT_DATA_HASH_SIZE, SCRIPT_HASH_SIZE, ScriptDataHash
from pycardano.plutus import COST_MODELS, CostModels, Datum, RedeemerMap, Redeemers
from pycardano.serialization import NonEmptyOrderedSet, default_encoder
from pycardano.transaction import MultiAsset, TransactionOutput, Value

__all__ = [
    "fee",
    "max_tx_fee",
    "bundle_size",
    "min_lovelace",
    "min_lovelace_pre_alonzo",
    "min_lovelace_post_alonzo",
    "script_data_hash",
    "tiered_reference_script_fee",
    "greater_than_version",
]


def tiered_reference_script_fee(context: ChainContext, scripts_size: int) -> int:
    """Calculate fee for reference scripts.

    Args:
        context (ChainContext): A chain context.
        scripts_size (int): Size of reference scripts in bytes.

    Returns:
        int: Fee for reference scripts.

    Raises:
        ValueError: If scripts size exceeds maximum allowed size
    """
    if (
        context.protocol_param.maximum_reference_scripts_size is None
        or context.protocol_param.min_fee_reference_scripts is None
    ):
        return 0

    max_size = context.protocol_param.maximum_reference_scripts_size["bytes"]
    if scripts_size > max_size:
        raise ValueError(
            f"Reference scripts size: {scripts_size} exceeds maximum allowed size ({max_size})."
        )

    total = 0.0
    if scripts_size:
        b = context.protocol_param.min_fee_reference_scripts["base"]
        r = math.ceil(context.protocol_param.min_fee_reference_scripts["range"])
        m = context.protocol_param.min_fee_reference_scripts["multiplier"]

        while scripts_size > r:
            total += b * r
            scripts_size = scripts_size - r
            b = b * m

        total += b * scripts_size

    return math.ceil(total)


def fee(
    context: ChainContext,
    length: int,
    exec_steps: int = 0,
    max_mem_unit: int = 0,
    ref_script_size: int = 0,
) -> int:
    """Calculate fee based on the length of a transaction's CBOR bytes and script execution.

    Args:
        context (ChainContext): A chain context.
        length (int): The length of CBOR bytes, which could usually be derived
            by `len(tx.to_cbor())`.
        exec_steps (int): Number of execution steps run by plutus scripts in the transaction.
        max_mem_unit (int): Max numer of memory units run by plutus scripts in the transaction.
        ref_script_size (int): Size of referenced scripts in the transaction.

    Return:
        int: Minimum acceptable transaction fee.
    """
    return int(
        math.ceil(length * context.protocol_param.min_fee_coefficient)
        + math.ceil(context.protocol_param.min_fee_constant)
        + math.ceil(exec_steps * context.protocol_param.price_step)
        + math.ceil(max_mem_unit * context.protocol_param.price_mem)
        + tiered_reference_script_fee(context, ref_script_size)
    )


def max_tx_fee(context: ChainContext, ref_script_size: int = 0) -> int:
    """Calculate the maximum possible transaction fee based on protocol parameters.

    Args:
        context (ChainContext): A chain context.
        ref_script_size (int): Size of reference scripts in the transaction.

    Returns:
        int: Maximum possible tx fee in lovelace.
    """
    return fee(
        context,
        context.protocol_param.max_tx_size,
        context.protocol_param.max_tx_ex_steps,
        context.protocol_param.max_tx_ex_mem,
        ref_script_size,
    )


def bundle_size(multi_asset: MultiAsset) -> int:
    """Calculate size of a multi-asset in words. (1 word = 8 bytes)

    Args:
        multi_asset (MultiAsset): Input multi asset.

    Returns:
        int: Number of words.
    """
    num_policies = len(multi_asset)
    num_assets = 0
    total_asset_name_len = 0

    # Only unique asset names are counted
    # see GitHub issue: https://github.com/Emurgo/cardano-serialization-lib/issues/194
    unique_assets = set()
    for p in multi_asset:
        num_assets += len(multi_asset[p])
        for n in multi_asset[p]:
            if n.payload not in unique_assets:
                unique_assets.add(n.payload)
                total_asset_name_len += len(n.payload)

    byte_len = num_assets * 12 + total_asset_name_len + num_policies * SCRIPT_HASH_SIZE
    return 6 + (byte_len + 7) // 8


def min_lovelace(
    context: ChainContext,
    output: Optional[TransactionOutput] = None,
    amount: Optional[Union[int, Value]] = None,
    has_datum: bool = False,
) -> int:
    """Calculate minimum lovelace a transaction output needs to hold.

    More info could be found in
    `this <https://github.com/input-output-hk/cardano-ledger/blob/master/doc/explanations/min-utxo-alonzo.rst>`_ page.

    Args:
        context (ChainContext): A chain context.
        output (TransactionOutput): A transaction output (for post-alonzo transactions).
        amount (Union[int, Value]): Amount from a transaction output (for pre-alonzo transactions).
        has_datum (bool): Whether the transaction output contains datum hash (for pre-alonzo transactions).

    Returns:
        int: Minimum required lovelace amount for this transaction output.
    """
    if output:
        return min_lovelace_post_alonzo(output, context)
    else:
        return min_lovelace_pre_alonzo(amount, context, has_datum)


def min_lovelace_pre_alonzo(
    amount: Union[int, Value, None], context: ChainContext, has_datum: bool = False
) -> int:
    """Calculate minimum lovelace a transaction output needs to hold.

    More info could be found in
    `this <https://github.com/input-output-hk/cardano-ledger/blob/master/doc/explanations/min-utxo-alonzo.rst>`_ page.

    Args:
        amount (Union[int, Value]): Amount from a transaction output.
        context (ChainContext): A chain context.
        has_datum (bool): Whether the transaction output contains datum hash.

    Returns:
        int: Minimum required lovelace amount for this transaction output.
    """
    if amount is None or isinstance(amount, int) or not amount.multi_asset:
        return context.protocol_param.min_utxo or 1_000_000

    b_size = bundle_size(amount.multi_asset)
    utxo_entry_size = 27
    data_hash_size = 10 if has_datum else 0
    finalized_size = utxo_entry_size + b_size + data_hash_size

    return finalized_size * context.protocol_param.coins_per_utxo_word


def min_lovelace_post_alonzo(output: TransactionOutput, context: ChainContext) -> int:
    """Calculate minimum lovelace a transaction output needs to hold post alonzo.

    This implementation is copied from the original Haskell implementation:
    https://github.com/input-output-hk/cardano-ledger/blob/eb053066c1d3bb51fb05978eeeab88afc0b049b2/eras/babbage/impl/src/Cardano/Ledger/Babbage/Rules/Utxo.hs#L242-L265

    Args:
        output (TransactionOutput): A transaction output.
        context (ChainContext): A chain context.

    Returns:
        int: Minimum required lovelace amount for this transaction output.
    """
    constant_overhead = 160

    amt = output.amount

    # If the amount of ADA is 0, a default value of 1 ADA will be used
    if amt.coin == 0:
        amt.coin = 1000000

    # Make sure we are using post-alonzo output
    tmp_out = TransactionOutput(
        output.address,
        output.amount,
        output.datum_hash,
        output.datum,
        output.script,
        True,
    )

    return (
        constant_overhead + len(tmp_out.to_cbor())
    ) * context.protocol_param.coins_per_utxo_byte


def script_data_hash(
    redeemers: Optional[Redeemers] = None,
    datums: Optional[Union[List[Datum], NonEmptyOrderedSet[Datum]]] = None,
    cost_models: Optional[Union[CostModels, Dict]] = None,
) -> ScriptDataHash:
    """Calculate plutus script data hash

    Args:
        redeemers (Optional[Redeemers]): Redeemers to include.
        datums (Optional[Union[List[Datum], NonEmptyOrderedSet[Datum]]]): Datums to include.
        cost_models (Optional[CostModels]): Cost models.

    Returns:
        ScriptDataHash: Plutus script data hash
    """
    if redeemers is None:
        redeemers = RedeemerMap()
        cost_models = {}
    elif len(redeemers) == 0:
        cost_models = {}
    elif not cost_models:
        cost_models = COST_MODELS

    redeemer_bytes = cbor2.dumps(redeemers, default=default_encoder)

    if datums:
        datum_bytes = cbor2.dumps(datums, default=default_encoder)
    else:
        datum_bytes = b""

    cost_models_bytes = cbor2.dumps(cost_models, default=default_encoder)

    return ScriptDataHash(
        blake2b(
            redeemer_bytes + datum_bytes + cost_models_bytes,
            SCRIPT_DATA_HASH_SIZE,
            encoder=RawEncoder,
        )
    )


def greater_than_version(version: Tuple[int, int]) -> bool:
    """Check if the current Python version is greater than or equal to the specified version

    Args:
        version (Tuple[int, int]): Tuple of major and minor version

    Returns:
        True if the current Python version is greater than or equal to the specified version
    """
    return sys.version_info >= version



import os
import tempfile
import time
import warnings
from fractions import Fraction
from typing import Dict, List, Optional, Union

from blockfrost import ApiError, ApiUrls, BlockFrostApi
from blockfrost.utils import Namespace

from pycardano.address import Address
from pycardano.backend.base import (
    ALONZO_COINS_PER_UTXO_WORD,
    ChainContext,
    GenesisParameters,
    ProtocolParameters,
)
from pycardano.cbor import cbor2
from pycardano.exception import TransactionFailedException
from pycardano.hash import SCRIPT_HASH_SIZE, DatumHash, ScriptHash
from pycardano.nativescript import NativeScript
from pycardano.network import Network
from pycardano.plutus import ExecutionUnits, PlutusScript, ScriptType, script_hash
from pycardano.serialization import RawCBOR
from pycardano.transaction import (
    Asset,
    AssetName,
    MultiAsset,
    TransactionInput,
    TransactionOutput,
    UTxO,
    Value,
)
from pycardano.types import JsonDict

__all__ = ["BlockFrostChainContext"]


def _try_fix_script(scripth: str, script: PlutusScript) -> PlutusScript:
    if str(script_hash(script)) == scripth:
        return script
    else:
        new_script = script.__class__(cbor2.loads(script))
        if str(script_hash(new_script)) == scripth:
            return new_script
        else:
            raise ValueError("Cannot recover script from hash.")


class BlockFrostChainContext(ChainContext):
    """A `BlockFrost <https://blockfrost.io/>`_ API wrapper for the client code to interact with.

    Args:
        project_id (str): A BlockFrost project ID obtained from https://blockfrost.io.
        network (Network): Network to use.
        base_url (str): Base URL for the BlockFrost API. Defaults to the preprod url.
    """

    api: BlockFrostApi
    _epoch_info: Namespace
    _epoch: Optional[int] = None
    _genesis_param: Optional[GenesisParameters] = None
    _protocol_param: Optional[ProtocolParameters] = None

    def __init__(
        self,
        project_id: str,
        network: Optional[Network] = None,
        base_url: Optional[str] = None,
    ):
        if network is not None:
            warnings.warn(
                "`network` argument will be deprecated in the future. Directly passing `base_url` is recommended."
            )
            self._network = network
        else:
            self._network = Network.TESTNET

        self._project_id = project_id
        self._base_url = (
            base_url
            if base_url
            else (
                ApiUrls.preprod.value
                if self.network == Network.TESTNET
                else ApiUrls.mainnet.value
            )
        )

        # Set network value to mainnet if base_url contains "mainnet".
        if "mainnet" in self._base_url:
            self._network = Network.MAINNET

        self.api = BlockFrostApi(project_id=self._project_id, base_url=self._base_url)
        self._epoch_info = self.api.epoch_latest()
        self._epoch = None
        self._genesis_param = None
        self._protocol_param = None

    def _check_epoch_and_update(self):
        if int(time.time()) >= self._epoch_info.end_time:
            self._epoch_info = self.api.epoch_latest()
            return True
        else:
            return False

    @property
    def network(self) -> Network:
        return self._network

    @property
    def epoch(self) -> int:
        if not self._epoch or self._check_epoch_and_update():
            new_epoch: int = self.api.epoch_latest().epoch
            self._epoch = new_epoch
        return self._epoch

    @property
    def last_block_slot(self) -> int:
        block = self.api.block_latest()
        return block.slot

    @property
    def genesis_param(self) -> GenesisParameters:
        if not self._genesis_param or self._check_epoch_and_update():
            params = vars(self.api.genesis())
            self._genesis_param = GenesisParameters(**params)
        return self._genesis_param

    @property
    def protocol_param(self) -> ProtocolParameters:
        if not self._protocol_param or self._check_epoch_and_update():
            params = self.api.epoch_latest_parameters()
            self._protocol_param = ProtocolParameters(
                min_fee_constant=int(params.min_fee_b),
                min_fee_coefficient=int(params.min_fee_a),
                max_block_size=int(params.max_block_size),
                max_tx_size=int(params.max_tx_size),
                max_block_header_size=int(params.max_block_header_size),
                key_deposit=int(params.key_deposit),
                pool_deposit=int(params.pool_deposit),
                pool_influence=Fraction(params.a0),
                monetary_expansion=Fraction(params.rho),
                treasury_expansion=Fraction(params.tau),
                decentralization_param=Fraction(params.decentralisation_param),
                extra_entropy=params.extra_entropy,
                protocol_major_version=int(params.protocol_major_ver),
                protocol_minor_version=int(params.protocol_minor_ver),
                min_utxo=int(params.min_utxo),
                min_pool_cost=int(params.min_pool_cost),
                price_mem=Fraction(params.price_mem),
                price_step=Fraction(params.price_step),
                max_tx_ex_mem=int(params.max_tx_ex_mem),
                max_tx_ex_steps=int(params.max_tx_ex_steps),
                max_block_ex_mem=int(params.max_block_ex_mem),
                max_block_ex_steps=int(params.max_block_ex_steps),
                max_val_size=int(params.max_val_size),
                collateral_percent=int(params.collateral_percent),
                max_collateral_inputs=int(params.max_collateral_inputs),
                coins_per_utxo_word=int(params.coins_per_utxo_word)
                or ALONZO_COINS_PER_UTXO_WORD,
                coins_per_utxo_byte=int(params.coins_per_utxo_size),
                cost_models={
                    k: v.to_dict() for k, v in params.cost_models.to_dict().items()
                },
                maximum_reference_scripts_size={"bytes": 200000},
                min_fee_reference_scripts={
                    "base": params.min_fee_ref_script_cost_per_byte,
                    "range": 200000,
                    "multiplier": 1,
                },
            )
        return self._protocol_param

    def _get_script(self, script_hash: str) -> ScriptType:
        script_type = self.api.script(script_hash).type
        if script_type.lower().startswith("plutusv"):
            ps = PlutusScript.from_version(
                int(script_type[-1]),
                bytes.fromhex(self.api.script_cbor(script_hash).cbor),
            )
            return _try_fix_script(script_hash, ps)
        else:
            script_json: JsonDict = self.api.script_json(
                script_hash, return_type="json"
            )["json"]
            return NativeScript.from_dict(script_json)

    def _utxos(self, address: str) -> List[UTxO]:
        try:
            results = self.api.address_utxos(address, gather_pages=True)
        except ApiError as e:
            if e.status_code == 404:
                return []
            else:
                raise e

        utxos = []

        for result in results:
            tx_in = TransactionInput.from_primitive(
                [result.tx_hash, result.output_index]
            )
            amount = result.amount
            lovelace_amount = 0
            multi_assets = MultiAsset()
            for item in amount:
                if item.unit == "lovelace":
                    lovelace_amount = int(item.quantity)
                else:
                    # The utxo contains Multi-asset
                    data = bytes.fromhex(item.unit)
                    policy_id = ScriptHash(data[:SCRIPT_HASH_SIZE])
                    asset_name = AssetName(data[SCRIPT_HASH_SIZE:])

                    if policy_id not in multi_assets:
                        multi_assets[policy_id] = Asset()
                    multi_assets[policy_id][asset_name] = int(item.quantity)

            amount = Value(lovelace_amount, multi_assets)

            datum_hash = (
                DatumHash.from_primitive(result.data_hash)
                if result.data_hash and result.inline_datum is None
                else None
            )

            datum = None

            if hasattr(result, "inline_datum") and result.inline_datum is not None:
                datum = RawCBOR(bytes.fromhex(result.inline_datum))

            script = None

            if (
                hasattr(result, "reference_script_hash")
                and result.reference_script_hash
            ):
                script = self._get_script(result.reference_script_hash)

            tx_out = TransactionOutput(
                Address.from_primitive(address),
                amount=amount,
                datum_hash=datum_hash,
                datum=datum,
                script=script,
            )
            utxos.append(UTxO(tx_in, tx_out))

        return utxos

    def submit_tx_cbor(self, cbor: Union[bytes, str]) -> str:
        """Submit a transaction.

        Args:
            cbor (Union[bytes, str]): The serialized transaction to be submitted.

        Returns:
            str: The transaction hash.

        Raises:
            :class:`TransactionFailedException`: When fails to submit the transaction.
        """
        if isinstance(cbor, str):
            cbor = bytes.fromhex(cbor)
        with tempfile.NamedTemporaryFile(delete=False) as f:
            f.write(cbor)
        try:
            response = self.api.transaction_submit(f.name)
        except ApiError as e:
            os.remove(f.name)
            raise TransactionFailedException(
                f"Failed to submit transaction. Error code: {e.status_code}. Error message: {e.message}"
            ) from e
        os.remove(f.name)
        return response

    def evaluate_tx_cbor(self, cbor: Union[bytes, str]) -> Dict[str, ExecutionUnits]:
        """Evaluate execution units of a transaction.

        Args:
            cbor (Union[bytes, str]): The serialized transaction to be evaluated.

        Returns:
            Dict[str, ExecutionUnits]: A list of execution units calculated for each of the transaction's redeemers

        Raises:
            :class:`TransactionFailedException`: When fails to evaluate the transaction.
        """
        if isinstance(cbor, bytes):
            cbor = cbor.hex()
        with tempfile.NamedTemporaryFile(delete=False, mode="w") as f:
            f.write(cbor)

        result = self.api.transaction_evaluate(f.name)
        os.remove(f.name)
        if not hasattr(result, "result"):
            raise TransactionFailedException(result)
        else:
            result = result.result
        return_val = {}
        if not hasattr(result, "EvaluationResult"):
            raise TransactionFailedException(result)
        else:
            for k in vars(result.EvaluationResult):
                return_val[k] = ExecutionUnits(
                    getattr(result.EvaluationResult, k).memory,
                    getattr(result.EvaluationResult, k).steps,
                )
            return return_val



"""Plutus related classes and functions."""

from __future__ import annotations

import inspect
import json
import typing
from dataclasses import dataclass, field, fields
from enum import Enum
from hashlib import sha256
from typing import Any, List, Optional, Type, Union

from cbor2 import CBORTag
from nacl.encoding import RawEncoder
from nacl.hash import blake2b
from typeguard import typechecked

from pycardano.cbor import cbor2
from pycardano.exception import DeserializeException, InvalidArgumentException
from pycardano.hash import DATUM_HASH_SIZE, SCRIPT_HASH_SIZE, DatumHash, ScriptHash
from pycardano.nativescript import NativeScript
from pycardano.serialization import (
    ArrayCBORSerializable,
    ByteString,
    CBORSerializable,
    DictCBORSerializable,
    IndefiniteList,
    Primitive,
    RawCBOR,
    default_encoder,
    limit_primitive_type,
)

__all__ = [
    "CostModels",
    "PLUTUS_V1_COST_MODEL",
    "PLUTUS_V2_COST_MODEL",
    "COST_MODELS",
    "PlutusData",
    "Datum",
    "RedeemerTag",
    "ExecutionUnits",
    "PlutusScript",
    "PlutusV1Script",
    "PlutusV2Script",
    "PlutusV3Script",
    "RawPlutusData",
    "Redeemer",
    "RedeemerKey",
    "RedeemerValue",
    "RedeemerMap",
    "Redeemers",
    "ScriptType",
    "datum_hash",
    "plutus_script_hash",
    "script_hash",
    "Unit",
]


# taken from https://stackoverflow.com/a/13624858
class classproperty(property):
    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


class CostModels(DictCBORSerializable):
    KEY_TYPE = int
    VALUE_TYPE = dict

    def to_shallow_primitive(self) -> dict:
        result: dict[bytes, Union[typing.List[Any], bytes]] = {}
        for language in sorted(self.keys()):
            cost_model = self[language]
            if language == 0:
                # Due to a bug in the Haskell implementation of ledger, we need to serialize the cost models twice.
                # See:
                # https://github.com/input-output-hk/cardano-ledger/blob/c9512ec56cd9b9ea20adea567649410289da0acc/eras/alonzo/test-suite/cddl-files/alonzo.cddl#L111-L115
                # https://github.com/input-output-hk/cardano-ledger/issues/2512
                l_cbor = cbor2.dumps(language, default=default_encoder)
                cm = IndefiniteList([cost_model[k] for k in sorted(cost_model.keys())])
                result[l_cbor] = cbor2.dumps(cm, default=default_encoder)
            else:
                result[language] = [cost_model[k] for k in cost_model.keys()]
        return result

    @classmethod
    @limit_primitive_type(dict)
    def from_primitive(cls: Type[CostModels], value: dict) -> CostModels:
        raise DeserializeException(
            "Deserialization of cost model is impossible, because some information is lost "
            "during serialization."
        )


# Copied from https://github.com/input-output-hk/cardano-configurations/blob/26b6b6de73f90e4777602b372798bf77addcc321/
# network/mainnet/genesis/alonzo.json#L27-L194
PLUTUS_V1_COST_MODEL = {
    "sha2_256-memory-arguments": 4,
    "equalsString-cpu-arguments-constant": 1000,
    "cekDelayCost-exBudgetMemory": 100,
    "lessThanEqualsByteString-cpu-arguments-intercept": 103599,
    "divideInteger-memory-arguments-minimum": 1,
    "appendByteString-cpu-arguments-slope": 621,
    "blake2b-cpu-arguments-slope": 29175,
    "iData-cpu-arguments": 150000,
    "encodeUtf8-cpu-arguments-slope": 1000,
    "unBData-cpu-arguments": 150000,
    "multiplyInteger-cpu-arguments-intercept": 61516,
    "cekConstCost-exBudgetMemory": 100,
    "nullList-cpu-arguments": 150000,
    "equalsString-cpu-arguments-intercept": 150000,
    "trace-cpu-arguments": 150000,
    "mkNilData-memory-arguments": 32,
    "lengthOfByteString-cpu-arguments": 150000,
    "cekBuiltinCost-exBudgetCPU": 29773,
    "bData-cpu-arguments": 150000,
    "subtractInteger-cpu-arguments-slope": 0,
    "unIData-cpu-arguments": 150000,
    "consByteString-memory-arguments-intercept": 0,
    "divideInteger-memory-arguments-slope": 1,
    "divideInteger-cpu-arguments-model-arguments-slope": 118,
    "listData-cpu-arguments": 150000,
    "headList-cpu-arguments": 150000,
    "chooseData-memory-arguments": 32,
    "equalsInteger-cpu-arguments-intercept": 136542,
    "sha3_256-cpu-arguments-slope": 82363,
    "sliceByteString-cpu-arguments-slope": 5000,
    "unMapData-cpu-arguments": 150000,
    "lessThanInteger-cpu-arguments-intercept": 179690,
    "mkCons-cpu-arguments": 150000,
    "appendString-memory-arguments-intercept": 0,
    "modInteger-cpu-arguments-model-arguments-slope": 118,
    "ifThenElse-cpu-arguments": 1,
    "mkNilPairData-cpu-arguments": 150000,
    "lessThanEqualsInteger-cpu-arguments-intercept": 145276,
    "addInteger-memory-arguments-slope": 1,
    "chooseList-memory-arguments": 32,
    "constrData-memory-arguments": 32,
    "decodeUtf8-cpu-arguments-intercept": 150000,
    "equalsData-memory-arguments": 1,
    "subtractInteger-memory-arguments-slope": 1,
    "appendByteString-memory-arguments-intercept": 0,
    "lengthOfByteString-memory-arguments": 4,
    "headList-memory-arguments": 32,
    "listData-memory-arguments": 32,
    "consByteString-cpu-arguments-intercept": 150000,
    "unIData-memory-arguments": 32,
    "remainderInteger-memory-arguments-minimum": 1,
    "bData-memory-arguments": 32,
    "lessThanByteString-cpu-arguments-slope": 248,
    "encodeUtf8-memory-arguments-intercept": 0,
    "cekStartupCost-exBudgetCPU": 100,
    "multiplyInteger-memory-arguments-intercept": 0,
    "unListData-memory-arguments": 32,
    "remainderInteger-cpu-arguments-model-arguments-slope": 118,
    "cekVarCost-exBudgetCPU": 29773,
    "remainderInteger-memory-arguments-slope": 1,
    "cekForceCost-exBudgetCPU": 29773,
    "sha2_256-cpu-arguments-slope": 29175,
    "equalsInteger-memory-arguments": 1,
    "indexByteString-memory-arguments": 1,
    "addInteger-memory-arguments-intercept": 1,
    "chooseUnit-cpu-arguments": 150000,
    "sndPair-cpu-arguments": 150000,
    "cekLamCost-exBudgetCPU": 29773,
    "fstPair-cpu-arguments": 150000,
    "quotientInteger-memory-arguments-minimum": 1,
    "decodeUtf8-cpu-arguments-slope": 1000,
    "lessThanInteger-memory-arguments": 1,
    "lessThanEqualsInteger-cpu-arguments-slope": 1366,
    "fstPair-memory-arguments": 32,
    "modInteger-memory-arguments-intercept": 0,
    "unConstrData-cpu-arguments": 150000,
    "lessThanEqualsInteger-memory-arguments": 1,
    "chooseUnit-memory-arguments": 32,
    "sndPair-memory-arguments": 32,
    "addInteger-cpu-arguments-intercept": 197209,
    "decodeUtf8-memory-arguments-slope": 8,
    "equalsData-cpu-arguments-intercept": 150000,
    "mapData-cpu-arguments": 150000,
    "mkPairData-cpu-arguments": 150000,
    "quotientInteger-cpu-arguments-constant": 148000,
    "consByteString-memory-arguments-slope": 1,
    "cekVarCost-exBudgetMemory": 100,
    "indexByteString-cpu-arguments": 150000,
    "unListData-cpu-arguments": 150000,
    "equalsInteger-cpu-arguments-slope": 1326,
    "cekStartupCost-exBudgetMemory": 100,
    "subtractInteger-cpu-arguments-intercept": 197209,
    "divideInteger-cpu-arguments-model-arguments-intercept": 425507,
    "divideInteger-memory-arguments-intercept": 0,
    "cekForceCost-exBudgetMemory": 100,
    "blake2b-cpu-arguments-intercept": 2477736,
    "remainderInteger-cpu-arguments-constant": 148000,
    "tailList-cpu-arguments": 150000,
    "encodeUtf8-cpu-arguments-intercept": 150000,
    "equalsString-cpu-arguments-slope": 1000,
    "lessThanByteString-memory-arguments": 1,
    "multiplyInteger-cpu-arguments-slope": 11218,
    "appendByteString-cpu-arguments-intercept": 396231,
    "lessThanEqualsByteString-cpu-arguments-slope": 248,
    "modInteger-memory-arguments-slope": 1,
    "addInteger-cpu-arguments-slope": 0,
    "equalsData-cpu-arguments-slope": 10000,
    "decodeUtf8-memory-arguments-intercept": 0,
    "chooseList-cpu-arguments": 150000,
    "constrData-cpu-arguments": 150000,
    "equalsByteString-memory-arguments": 1,
    "cekApplyCost-exBudgetCPU": 29773,
    "quotientInteger-memory-arguments-slope": 1,
    "verifySignature-cpu-arguments-intercept": 3345831,
    "unMapData-memory-arguments": 32,
    "mkCons-memory-arguments": 32,
    "sliceByteString-memory-arguments-slope": 1,
    "sha3_256-memory-arguments": 4,
    "ifThenElse-memory-arguments": 1,
    "mkNilPairData-memory-arguments": 32,
    "equalsByteString-cpu-arguments-slope": 247,
    "appendString-cpu-arguments-intercept": 150000,
    "quotientInteger-cpu-arguments-model-arguments-slope": 118,
    "cekApplyCost-exBudgetMemory": 100,
    "equalsString-memory-arguments": 1,
    "multiplyInteger-memory-arguments-slope": 1,
    "cekBuiltinCost-exBudgetMemory": 100,
    "remainderInteger-memory-arguments-intercept": 0,
    "sha2_256-cpu-arguments-intercept": 2477736,
    "remainderInteger-cpu-arguments-model-arguments-intercept": 425507,
    "lessThanEqualsByteString-memory-arguments": 1,
    "tailList-memory-arguments": 32,
    "mkNilData-cpu-arguments": 150000,
    "chooseData-cpu-arguments": 150000,
    "unBData-memory-arguments": 32,
    "blake2b-memory-arguments": 4,
    "iData-memory-arguments": 32,
    "nullList-memory-arguments": 32,
    "cekDelayCost-exBudgetCPU": 29773,
    "subtractInteger-memory-arguments-intercept": 1,
    "lessThanByteString-cpu-arguments-intercept": 103599,
    "consByteString-cpu-arguments-slope": 1000,
    "appendByteString-memory-arguments-slope": 1,
    "trace-memory-arguments": 32,
    "divideInteger-cpu-arguments-constant": 148000,
    "cekConstCost-exBudgetCPU": 29773,
    "encodeUtf8-memory-arguments-slope": 8,
    "quotientInteger-cpu-arguments-model-arguments-intercept": 425507,
    "mapData-memory-arguments": 32,
    "appendString-cpu-arguments-slope": 1000,
    "modInteger-cpu-arguments-constant": 148000,
    "verifySignature-cpu-arguments-slope": 1,
    "unConstrData-memory-arguments": 32,
    "quotientInteger-memory-arguments-intercept": 0,
    "equalsByteString-cpu-arguments-constant": 150000,
    "sliceByteString-memory-arguments-intercept": 0,
    "mkPairData-memory-arguments": 32,
    "equalsByteString-cpu-arguments-intercept": 112536,
    "appendString-memory-arguments-slope": 1,
    "lessThanInteger-cpu-arguments-slope": 497,
    "modInteger-cpu-arguments-model-arguments-intercept": 425507,
    "modInteger-memory-arguments-minimum": 1,
    "sha3_256-cpu-arguments-intercept": 0,
    "verifySignature-memory-arguments": 1,
    "cekLamCost-exBudgetMemory": 100,
    "sliceByteString-cpu-arguments-intercept": 150000,
}


PLUTUS_V2_COST_MODEL = {
    "addInteger-cpu-arguments-intercept": 205665,
    "addInteger-cpu-arguments-slope": 812,
    "addInteger-memory-arguments-intercept": 1,
    "addInteger-memory-arguments-slope": 1,
    "appendByteString-cpu-arguments-intercept": 1000,
    "appendByteString-cpu-arguments-slope": 571,
    "appendByteString-memory-arguments-intercept": 0,
    "appendByteString-memory-arguments-slope": 1,
    "appendString-cpu-arguments-intercept": 1000,
    "appendString-cpu-arguments-slope": 24177,
    "appendString-memory-arguments-intercept": 4,
    "appendString-memory-arguments-slope": 1,
    "bData-cpu-arguments": 1000,
    "bData-memory-arguments": 32,
    "blake2b_256-cpu-arguments-intercept": 117366,
    "blake2b_256-cpu-arguments-slope": 10475,
    "blake2b_256-memory-arguments": 4,
    "cekApplyCost-exBudgetCPU": 23000,
    "cekApplyCost-exBudgetMemory": 100,
    "cekBuiltinCost-exBudgetCPU": 23000,
    "cekBuiltinCost-exBudgetMemory": 100,
    "cekConstCost-exBudgetCPU": 23000,
    "cekConstCost-exBudgetMemory": 100,
    "cekDelayCost-exBudgetCPU": 23000,
    "cekDelayCost-exBudgetMemory": 100,
    "cekForceCost-exBudgetCPU": 23000,
    "cekForceCost-exBudgetMemory": 100,
    "cekLamCost-exBudgetCPU": 23000,
    "cekLamCost-exBudgetMemory": 100,
    "cekStartupCost-exBudgetCPU": 100,
    "cekStartupCost-exBudgetMemory": 100,
    "cekVarCost-exBudgetCPU": 23000,
    "cekVarCost-exBudgetMemory": 100,
    "chooseData-cpu-arguments": 19537,
    "chooseData-memory-arguments": 32,
    "chooseList-cpu-arguments": 175354,
    "chooseList-memory-arguments": 32,
    "chooseUnit-cpu-arguments": 46417,
    "chooseUnit-memory-arguments": 4,
    "consByteString-cpu-arguments-intercept": 221973,
    "consByteString-cpu-arguments-slope": 511,
    "consByteString-memory-arguments-intercept": 0,
    "consByteString-memory-arguments-slope": 1,
    "constrData-cpu-arguments": 89141,
    "constrData-memory-arguments": 32,
    "decodeUtf8-cpu-arguments-intercept": 497525,
    "decodeUtf8-cpu-arguments-slope": 14068,
    "decodeUtf8-memory-arguments-intercept": 4,
    "decodeUtf8-memory-arguments-slope": 2,
    "divideInteger-cpu-arguments-constant": 196500,
    "divideInteger-cpu-arguments-model-arguments-intercept": 453240,
    "divideInteger-cpu-arguments-model-arguments-slope": 220,
    "divideInteger-memory-arguments-intercept": 0,
    "divideInteger-memory-arguments-minimum": 1,
    "divideInteger-memory-arguments-slope": 1,
    "encodeUtf8-cpu-arguments-intercept": 1000,
    "encodeUtf8-cpu-arguments-slope": 28662,
    "encodeUtf8-memory-arguments-intercept": 4,
    "encodeUtf8-memory-arguments-slope": 2,
    "equalsByteString-cpu-arguments-constant": 245000,
    "equalsByteString-cpu-arguments-intercept": 216773,
    "equalsByteString-cpu-arguments-slope": 62,
    "equalsByteString-memory-arguments": 1,
    "equalsData-cpu-arguments-intercept": 1060367,
    "equalsData-cpu-arguments-slope": 12586,
    "equalsData-memory-arguments": 1,
    "equalsInteger-cpu-arguments-intercept": 208512,
    "equalsInteger-cpu-arguments-slope": 421,
    "equalsInteger-memory-arguments": 1,
    "equalsString-cpu-arguments-constant": 187000,
    "equalsString-cpu-arguments-intercept": 1000,
    "equalsString-cpu-arguments-slope": 52998,
    "equalsString-memory-arguments": 1,
    "fstPair-cpu-arguments": 80436,
    "fstPair-memory-arguments": 32,
    "headList-cpu-arguments": 43249,
    "headList-memory-arguments": 32,
    "iData-cpu-arguments": 1000,
    "iData-memory-arguments": 32,
    "ifThenElse-cpu-arguments": 80556,
    "ifThenElse-memory-arguments": 1,
    "indexByteString-cpu-arguments": 57667,
    "indexByteString-memory-arguments": 4,
    "lengthOfByteString-cpu-arguments": 1000,
    "lengthOfByteString-memory-arguments": 10,
    "lessThanByteString-cpu-arguments-intercept": 197145,
    "lessThanByteString-cpu-arguments-slope": 156,
    "lessThanByteString-memory-arguments": 1,
    "lessThanEqualsByteString-cpu-arguments-intercept": 197145,
    "lessThanEqualsByteString-cpu-arguments-slope": 156,
    "lessThanEqualsByteString-memory-arguments": 1,
    "lessThanEqualsInteger-cpu-arguments-intercept": 204924,
    "lessThanEqualsInteger-cpu-arguments-slope": 473,
    "lessThanEqualsInteger-memory-arguments": 1,
    "lessThanInteger-cpu-arguments-intercept": 208896,
    "lessThanInteger-cpu-arguments-slope": 511,
    "lessThanInteger-memory-arguments": 1,
    "listData-cpu-arguments": 52467,
    "listData-memory-arguments": 32,
    "mapData-cpu-arguments": 64832,
    "mapData-memory-arguments": 32,
    "mkCons-cpu-arguments": 65493,
    "mkCons-memory-arguments": 32,
    "mkNilData-cpu-arguments": 22558,
    "mkNilData-memory-arguments": 32,
    "mkNilPairData-cpu-arguments": 16563,
    "mkNilPairData-memory-arguments": 32,
    "mkPairData-cpu-arguments": 76511,
    "mkPairData-memory-arguments": 32,
    "modInteger-cpu-arguments-constant": 196500,
    "modInteger-cpu-arguments-model-arguments-intercept": 453240,
    "modInteger-cpu-arguments-model-arguments-slope": 220,
    "modInteger-memory-arguments-intercept": 0,
    "modInteger-memory-arguments-minimum": 1,
    "modInteger-memory-arguments-slope": 1,
    "multiplyInteger-cpu-arguments-intercept": 69522,
    "multiplyInteger-cpu-arguments-slope": 11687,
    "multiplyInteger-memory-arguments-intercept": 0,
    "multiplyInteger-memory-arguments-slope": 1,
    "nullList-cpu-arguments": 60091,
    "nullList-memory-arguments": 32,
    "quotientInteger-cpu-arguments-constant": 196500,
    "quotientInteger-cpu-arguments-model-arguments-intercept": 453240,
    "quotientInteger-cpu-arguments-model-arguments-slope": 220,
    "quotientInteger-memory-arguments-intercept": 0,
    "quotientInteger-memory-arguments-minimum": 1,
    "quotientInteger-memory-arguments-slope": 1,
    "remainderInteger-cpu-arguments-constant": 196500,
    "remainderInteger-cpu-arguments-model-arguments-intercept": 453240,
    "remainderInteger-cpu-arguments-model-arguments-slope": 220,
    "remainderInteger-memory-arguments-intercept": 0,
    "remainderInteger-memory-arguments-minimum": 1,
    "remainderInteger-memory-arguments-slope": 1,
    "serialiseData-cpu-arguments-intercept": 1159724,
    "serialiseData-cpu-arguments-slope": 392670,
    "serialiseData-memory-arguments-intercept": 0,
    "serialiseData-memory-arguments-slope": 2,
    "sha2_256-cpu-arguments-intercept": 806990,
    "sha2_256-cpu-arguments-slope": 30482,
    "sha2_256-memory-arguments": 4,
    "sha3_256-cpu-arguments-intercept": 1927926,
    "sha3_256-cpu-arguments-slope": 82523,
    "sha3_256-memory-arguments": 4,
    "sliceByteString-cpu-arguments-intercept": 265318,
    "sliceByteString-cpu-arguments-slope": 0,
    "sliceByteString-memory-arguments-intercept": 4,
    "sliceByteString-memory-arguments-slope": 0,
    "sndPair-cpu-arguments": 85931,
    "sndPair-memory-arguments": 32,
    "subtractInteger-cpu-arguments-intercept": 205665,
    "subtractInteger-cpu-arguments-slope": 812,
    "subtractInteger-memory-arguments-intercept": 1,
    "subtractInteger-memory-arguments-slope": 1,
    "tailList-cpu-arguments": 41182,
    "tailList-memory-arguments": 32,
    "trace-cpu-arguments": 212342,
    "trace-memory-arguments": 32,
    "unBData-cpu-arguments": 31220,
    "unBData-memory-arguments": 32,
    "unConstrData-cpu-arguments": 32696,
    "unConstrData-memory-arguments": 32,
    "unIData-cpu-arguments": 43357,
    "unIData-memory-arguments": 32,
    "unListData-cpu-arguments": 32247,
    "unListData-memory-arguments": 32,
    "unMapData-cpu-arguments": 38314,
    "unMapData-memory-arguments": 32,
    "verifyEcdsaSecp256k1Signature-cpu-arguments": 20000000000,
    "verifyEcdsaSecp256k1Signature-memory-arguments": 20000000000,
    "verifyEd25519Signature-cpu-arguments-intercept": 9462713,
    "verifyEd25519Signature-cpu-arguments-slope": 1021,
    "verifyEd25519Signature-memory-arguments": 10,
    "verifySchnorrSecp256k1Signature-cpu-arguments-intercept": 20000000000,
    "verifySchnorrSecp256k1Signature-cpu-arguments-slope": 0,
    "verifySchnorrSecp256k1Signature-memory-arguments": 20000000000,
}

COST_MODELS = CostModels({0: PLUTUS_V1_COST_MODEL})
"""A dictionary of current cost models, which could be used to calculate script data hash."""


def get_tag(constr_id: int) -> Optional[int]:
    if 0 <= constr_id < 7:
        return 121 + constr_id
    elif 7 <= constr_id < 128:
        return 1280 + (constr_id - 7)
    else:
        return None


def get_constructor_id_and_fields(
    raw_tag: CBORTag,
) -> typing.Tuple[int, typing.List[Any]]:
    tag = raw_tag.tag
    if tag == 102:
        if len(raw_tag.value) != 2:
            raise DeserializeException(
                f"Expect the length of value to be exactly 2, got {len(raw_tag.value)} instead."
            )
        return raw_tag.value[0], raw_tag.value[1]
    else:
        if 121 <= tag < 128:
            constr = tag - 121
        elif 1280 <= tag < 1536:
            constr = tag - 1280 + 7
        else:
            raise DeserializeException(f"Unexpected tag for RawPlutusData: {tag}")
        return constr, raw_tag.value


def id_map(cls, skip_constructor=False):
    """
    Constructs a unique representation of a PlutusData type definition.
    Intended for automatic constructor generation.
    """
    if cls == bytes or cls == ByteString:
        return "bytes"
    if cls == int:
        return "int"
    if cls == RawCBOR or cls == RawPlutusData or cls == Datum:
        return "any"
    if cls == IndefiniteList:
        return "list"
    if hasattr(cls, "__origin__"):
        origin = getattr(cls, "__origin__")
        if origin == list:
            prefix = "list"
        elif origin == dict:
            prefix = "map"
        elif origin == typing.Union:
            prefix = "union"
        else:
            raise TypeError(
                f"Unexpected parameterized type for automatic constructor generation: {cls}"
            )
        return prefix + "<" + ",".join(id_map(a) for a in cls.__args__) + ">"
    if issubclass(cls, PlutusData):
        return (
            "cons["
            + cls.__name__
            + "]("
            + (str(cls.CONSTR_ID) if not skip_constructor else "_")
            + ";"
            + ",".join(f.name + ":" + id_map(f.type) for f in fields(cls))
            + ")"
        )
    raise TypeError(f"Unexpected type for automatic constructor generation: {cls}")


@dataclass(repr=False)
class PlutusData(ArrayCBORSerializable):
    """
    PlutusData is a helper class that can serialize itself into a CBOR format, which could be intepreted as
    a data structure in Plutus scripts.
    It is not required to use this class to interact with Plutus scripts. However, wrapping datum in PlutusData
    class will reduce the complexity of serialization and deserialization tremendously.

    Examples:

        >>> @dataclass
        ... class Test(PlutusData):
        ...     CONSTR_ID = 1
        ...     a: int
        ...     b: bytes
        >>> test = Test(123, b"321")
        >>> test.to_cbor_hex()
        'd87a9f187b43333231ff'
        >>> assert test == Test.from_cbor("d87a9f187b43333231ff")
    """

    MAX_BYTES_SIZE = 64

    @classproperty
    def CONSTR_ID(cls):
        """
        Constructor ID of this plutus data.
        It is primarily used by Plutus core to reconstruct a data structure from serialized CBOR bytes.
        The default implementation is an almost unique, deterministic constructor ID in the range 1 - 2^32 based
        on class attributes, types and class name.
        """
        k = f"_CONSTR_ID_{cls.__name__}"
        if not hasattr(cls, k):
            det_string = id_map(cls, skip_constructor=True)
            det_hash = sha256(det_string.encode("utf8")).hexdigest()
            setattr(cls, k, int(det_hash, 16) % 2**32)

        return getattr(cls, k)

    def __post_init__(self):
        valid_types = (
            RawPlutusData,
            PlutusData,
            dict,
            IndefiniteList,
            int,
            ByteString,
            bytes,
        )
        for f in fields(self):
            if inspect.isclass(f.type) and not issubclass(f.type, valid_types):
                raise TypeError(
                    f"Invalid field type: {f.type}. A field in PlutusData should be one of {valid_types}"
                )

            data = getattr(self, f.name)
            if isinstance(data, bytes) and len(data) > 64:
                raise InvalidArgumentException(
                    f"The size of {data} exceeds {self.MAX_BYTES_SIZE} bytes. "
                    "Use pycardano.serialization.ByteString for long bytes."
                )

    def to_shallow_primitive(self) -> CBORTag:
        primitives: Primitive = super().to_shallow_primitive()
        if primitives:
            primitives = IndefiniteList(primitives)
        tag = get_tag(self.CONSTR_ID)
        if tag:
            return CBORTag(tag, primitives)
        else:
            return CBORTag(102, [self.CONSTR_ID, primitives])

    @classmethod
    @limit_primitive_type(CBORTag)
    def from_primitive(cls: Type[PlutusData], value: CBORTag) -> PlutusData:
        if value.tag == 102:
            tag = value.value[0]
            if tag != cls.CONSTR_ID:
                raise DeserializeException(
                    f"Unexpected constructor ID for {cls}. Expect {cls.CONSTR_ID}, got "
                    f"{tag} instead."
                )
            if len(value.value) != 2:
                raise DeserializeException(
                    f"Expect the length of value to be exactly 2, got {len(value.value)} instead."
                )
            return super(PlutusData, cls).from_primitive(value.value[1])
        else:
            expected_tag = get_tag(cls.CONSTR_ID)
            if expected_tag != value.tag:
                raise DeserializeException(
                    f"Unexpected constructor ID for {cls}. Expect {expected_tag}, got "
                    f"{value.tag} instead."
                )
            return super(PlutusData, cls).from_primitive(value.value)

    def hash(self) -> DatumHash:
        return datum_hash(self)

    def to_dict(self) -> dict:
        """
        Convert to a dictionary.

        Returns:
            str: a dict PlutusData that can be JSON encoded.
        """

        def _dfs(obj):
            """
            Reference of Haskell's implementation:
            https://github.com/input-output-hk/cardano-node/blob/baa9b5e59c5d448d475f94cc88a31a5857c2bda5/cardano-api/
            src/Cardano/Api/ScriptData.hs#L449-L474
            """
            if isinstance(obj, int):
                return {"int": obj}
            elif isinstance(obj, bytes):
                return {"bytes": obj.hex()}
            elif isinstance(obj, ByteString):
                return {"bytes": obj.value.hex()}
            elif isinstance(obj, IndefiniteList) or isinstance(obj, list):
                return {"list": [_dfs(item) for item in obj]}
            elif isinstance(obj, dict):
                return {"map": [{"v": _dfs(v), "k": _dfs(k)} for k, v in obj.items()]}
            elif isinstance(obj, PlutusData):
                return {
                    "constructor": obj.CONSTR_ID,
                    "fields": [_dfs(getattr(obj, f.name)) for f in fields(obj)],
                }
            elif isinstance(obj, RawPlutusData):
                return obj.to_dict()
            elif isinstance(obj, RawCBOR):
                return RawPlutusData.from_cbor(obj.cbor).to_dict()
            else:
                raise TypeError(f"Unexpected type {type(obj)}")

        return _dfs(self)

    def to_json(self, **kwargs) -> str:  # type: ignore
        """Convert to a json string

        Args:
            **kwargs: Extra key word arguments to be passed to `json.dumps()`

        Returns:
            str: a JSON encoded PlutusData.
        """

        return json.dumps(self.to_dict(), **kwargs)

    @classmethod
    def from_dict(cls: Type[PlutusData], data: dict) -> PlutusData:
        """Convert a dictionary to PlutusData

        Args:
            data (dict): A dictionary.

        Returns:
            PlutusData: Restored PlutusData.
        """

        def _dfs(obj):
            if isinstance(obj, dict):
                if "constructor" in obj:
                    if obj["constructor"] != cls.CONSTR_ID:
                        raise DeserializeException(
                            f"Mismatch between constructors in class {cls.__name__}, expect: {cls.CONSTR_ID}, "
                            f"got: {obj['constructor']} instead."
                        )
                    converted_fields = []
                    for f, f_info in zip(obj["fields"], fields(cls)):
                        if inspect.isclass(f_info.type) and issubclass(
                            f_info.type, PlutusData
                        ):
                            converted_fields.append(f_info.type.from_dict(f))
                        elif f_info.type == Datum:
                            converted_fields.append(RawPlutusData.from_dict(f))
                        elif (
                            hasattr(f_info.type, "__origin__")
                            and f_info.type.__origin__ is Union
                        ):
                            t_args = f_info.type.__args__
                            found_match = False
                            for t in t_args:
                                if (
                                    inspect.isclass(t)
                                    and issubclass(t, PlutusData)
                                    and t.CONSTR_ID == f["constructor"]
                                ):
                                    converted_fields.append(t.from_dict(f))
                                    found_match = True
                                    break
                            if not found_match:
                                raise DeserializeException(
                                    f"Unexpected data structure: {f}."
                                )
                        elif (
                            hasattr(f_info.type, "__origin__")
                            and f_info.type.__origin__ is list
                        ):
                            t_args = f_info.type.__args__
                            if len(t_args) != 1:
                                raise DeserializeException(
                                    f"List types need exactly one type argument, but got {t_args}"
                                )
                            if "list" not in f:
                                raise DeserializeException(
                                    f'Expected type "list" for constructor List but got {f}'
                                )
                            t = t_args[0]
                            if inspect.isclass(t) and issubclass(t, PlutusData):
                                converted_fields.append(t.from_dict(f))
                            else:
                                converted_fields.append(_dfs(f))

                        elif (
                            hasattr(f_info.type, "__origin__")
                            and f_info.type.__origin__ is dict
                        ):
                            t_args = f_info.type.__args__
                            if len(t_args) != 2:
                                raise DeserializeException(
                                    "Dict type with wrong number of arguments"
                                )
                            if "map" not in f:
                                raise DeserializeException(
                                    f'Expected type "map" in object but got "{f}"'
                                )
                            key_t = t_args[0]
                            val_t = t_args[1]
                            if inspect.isclass(key_t) and issubclass(key_t, PlutusData):
                                key_convert = key_t.from_dict
                            else:
                                key_convert = _dfs
                            if inspect.isclass(val_t) and issubclass(val_t, PlutusData):
                                val_convert = val_t.from_dict
                            else:
                                val_convert = _dfs
                            converted_fields.append(
                                {
                                    key_convert(pair["k"]): val_convert(pair["v"])
                                    for pair in f["map"]
                                }
                            )
                        else:
                            converted_fields.append(_dfs(f))
                    return cls(*converted_fields)
                elif "map" in obj:
                    return {_dfs(pair["k"]): _dfs(pair["v"]) for pair in obj["map"]}
                elif "int" in obj:
                    return obj["int"]
                elif "bytes" in obj:
                    if len(obj["bytes"]) > 64:
                        return ByteString(bytes.fromhex(obj["bytes"]))
                    else:
                        return bytes.fromhex(obj["bytes"])
                elif "list" in obj:
                    return IndefiniteList([_dfs(item) for item in obj["list"]])
                else:
                    raise DeserializeException(f"Unexpected data structure: {obj}")
            else:
                raise TypeError(f"Unexpected data type: {type(obj)}")

        return _dfs(data)

    @classmethod
    def from_json(cls: Type[PlutusData], data: str) -> PlutusData:
        """Restore a json encoded string to a PlutusData.

        Args:
            data (str): An encoded json string.

        Returns:
            PlutusData: The restored PlutusData.
        """
        obj = json.loads(data)
        return cls.from_dict(obj)

    def __deepcopy__(self, memo):
        return self.__class__.from_cbor(self.to_cbor_hex())


RawDatum = Union[PlutusData, dict, int, bytes, IndefiniteList, RawCBOR, CBORTag]


@dataclass(repr=True)
class RawPlutusData(CBORSerializable):
    data: RawDatum

    def to_primitive(self) -> Primitive:
        def _dfs(obj):
            if isinstance(obj, list) and obj:
                return IndefiniteList([_dfs(item) for item in obj])
            elif isinstance(obj, dict):
                return {_dfs(k): _dfs(v) for k, v in obj.items()}
            elif isinstance(obj, CBORTag) and isinstance(obj.value, list) and obj.value:
                if obj.tag != 102:
                    value = IndefiniteList([_dfs(item) for item in obj.value])
                else:
                    value = [_dfs(item) for item in obj.value]
                return CBORTag(tag=obj.tag, value=value)
            return obj

        return _dfs(self.data)

    def to_dict(self) -> dict:
        """
        Convert to a dictionary.

        Returns:
            str: a dict RawPlutusData that can be JSON encoded.
        """

        def _dfs(obj):
            if isinstance(obj, int):
                return {"int": obj}
            elif isinstance(obj, bytes):
                return {"bytes": obj.hex()}
            elif isinstance(obj, ByteString):
                return {"bytes": obj.value.hex()}
            elif isinstance(obj, IndefiniteList) or isinstance(obj, list):
                return {"list": [_dfs(item) for item in obj]}
            elif isinstance(obj, dict):
                return {"map": [{"v": _dfs(v), "k": _dfs(k)} for k, v in obj.items()]}
            elif isinstance(obj, CBORTag):
                constructor, fields = get_constructor_id_and_fields(obj)
                return {"constructor": constructor, "fields": [_dfs(f) for f in fields]}
            elif isinstance(obj, RawCBOR):
                return RawPlutusData.from_cbor(obj.cbor).to_dict()
            raise TypeError(f"Unexpected type {type(obj)}")

        return _dfs(RawPlutusData.to_primitive(self))

    def to_json(self, **kwargs) -> str:  # type: ignore
        """Convert to a json string

        Args:
            **kwargs: Extra key word arguments to be passed to `json.dumps()`

        Returns:
            str: a JSON encoded RawPlutusData.
        """

        return json.dumps(RawPlutusData.to_dict(self), **kwargs)

    @classmethod
    @limit_primitive_type(
        PlutusData, dict, int, bytes, IndefiniteList, RawCBOR, CBORTag
    )  # equal to RawDatum parameter list
    def from_primitive(cls: Type[RawPlutusData], value: RawDatum) -> RawPlutusData:
        return cls(value)

    @classmethod
    def from_dict(cls: Type[RawPlutusData], data: dict) -> RawPlutusData:
        """Convert a dictionary to RawPlutusData

        Args:
            data (dict): A dictionary.

        Returns:
            RawPlutusData: Restored RawPlutusData.
        """

        def _dfs(obj):
            if isinstance(obj, dict):
                if "constructor" in obj:
                    converted_fields = []
                    for f in obj["fields"]:
                        converted_fields.append(_dfs(f))
                    tag = get_tag(obj["constructor"])
                    if tag is None:
                        return CBORTag(
                            102, [obj["constructor"], IndefiniteList(converted_fields)]
                        )
                    else:
                        return CBORTag(tag, converted_fields)
                elif "map" in obj:
                    return {_dfs(pair["k"]): _dfs(pair["v"]) for pair in obj["map"]}
                elif "int" in obj:
                    return obj["int"]
                elif "bytes" in obj:
                    if len(obj["bytes"]) > 64:
                        return ByteString(bytes.fromhex(obj["bytes"]))
                    else:
                        return bytes.fromhex(obj["bytes"])
                elif "list" in obj:
                    return IndefiniteList([_dfs(item) for item in obj["list"]])
                else:
                    raise DeserializeException(f"Unexpected data structure: {obj}")
            else:
                raise TypeError(f"Unexpected data type: {type(obj)}")

        return cls(_dfs(data))

    @classmethod
    def from_json(cls: Type[RawPlutusData], data: str) -> RawPlutusData:
        """Restore a json encoded string to a RawPlutusData.

        Args:
            data (str): An encoded json string.

        Returns:
            RawPlutusData: The restored RawPlutusData.
        """
        obj = json.loads(data)
        return cls.from_dict(obj)

    def __deepcopy__(self, memo):
        return self.__class__.from_cbor(self.to_cbor_hex())


Datum = Union[PlutusData, dict, int, bytes, IndefiniteList, RawCBOR, RawPlutusData]
"""Plutus Datum type. A Union type that contains all valid datum types."""


def datum_hash(datum: Datum) -> DatumHash:
    return DatumHash(
        blake2b(
            cbor2.dumps(datum, default=default_encoder),
            DATUM_HASH_SIZE,
            encoder=RawEncoder,
        )
    )


class RedeemerTag(CBORSerializable, Enum):
    """
    Redeemer tag, which indicates the type of redeemer.
    """

    SPEND = 0
    MINT = 1
    CERTIFICATE = 2
    WITHDRAWAL = 3
    VOTING = 4
    PROPOSING = 5

    def to_primitive(self) -> int:
        return self.value

    @classmethod
    @limit_primitive_type(int)
    def from_primitive(cls: Type[RedeemerTag], value: int) -> RedeemerTag:
        return cls(value)


@dataclass(repr=False)
class ExecutionUnits(ArrayCBORSerializable):
    mem: int

    steps: int

    def __add__(self, other: ExecutionUnits) -> ExecutionUnits:
        if not isinstance(other, ExecutionUnits):
            raise TypeError(
                f"Expect type: {ExecutionUnits}, got {type(other)} instead."
            )
        return ExecutionUnits(self.mem + other.mem, self.steps + other.steps)

    def is_empty(self) -> bool:
        return self.mem == 0 and self.steps == 0

    def __bool__(self):
        return not self.is_empty()


@dataclass(repr=False)
class Redeemer(ArrayCBORSerializable):
    tag: Optional[RedeemerTag] = field(default=None, init=False)

    index: int = field(default=0, init=False)

    data: Any

    ex_units: Optional[ExecutionUnits] = None

    @classmethod
    @limit_primitive_type(list, tuple)
    def from_primitive(cls: Type[Redeemer], values: list) -> Redeemer:
        if isinstance(values[2], CBORTag) and cls is Redeemer:
            values[2] = RawPlutusData.from_primitive(values[2])
        redeemer = super(Redeemer, cls).from_primitive([values[2], values[3]])
        redeemer.tag = RedeemerTag.from_primitive(values[0])
        redeemer.index = values[1]
        return redeemer


@dataclass(repr=False)
class RedeemerKey(ArrayCBORSerializable):
    tag: RedeemerTag

    index: int = field(default=0)

    @classmethod
    @limit_primitive_type(list, tuple)
    def from_primitive(cls: Type[RedeemerKey], values: list) -> RedeemerKey:
        tag = RedeemerTag.from_primitive(values[0])
        index = values[1]
        return cls(tag, index)

    def __eq__(self, other):
        if not isinstance(other, RedeemerKey):
            return False
        return self.tag == other.tag and self.index == other.index

    def __hash__(self):
        return hash(self.to_cbor())


@dataclass(repr=False)
class RedeemerValue(ArrayCBORSerializable):
    data: Any

    ex_units: ExecutionUnits

    @classmethod
    @limit_primitive_type(list, tuple)
    def from_primitive(cls: Type[RedeemerValue], values: list) -> RedeemerValue:
        if isinstance(values[0], CBORTag) and cls is RedeemerValue:
            values[0] = RawPlutusData.from_primitive(values[0])
        return super(RedeemerValue, cls).from_primitive([values[0], values[1]])

    def __eq__(self, other):
        if not isinstance(other, RedeemerValue):
            return False
        return self.data == other.data and self.ex_units == other.ex_units


@typechecked
class RedeemerMap(DictCBORSerializable):
    KEY_TYPE = RedeemerKey

    VALUE_TYPE = RedeemerValue


Redeemers = Union[List[Redeemer], RedeemerMap]


def plutus_script_hash(script: Union[NativeScript, PlutusScript]) -> ScriptHash:
    """Calculates the hash of a Plutus script.

    Args:
        script (Union[bytes, PlutusScript]): A plutus script.

    Returns:
        ScriptHash: blake2b hash of the script.
    """
    return script_hash(script)


class PlutusScript(CBORSerializable, bytes):
    """
    Plutus script class.

    This class is a base class for all Plutus script versions.

    Example - Load a Plutus script from `test/resources/scriptV2.plutus <https://github.com/Python-Cardano/pycardano/blob/main/test/resources/scriptV2.plutus>`_ and get its address: # noqa: E501


        >>> from pycardano import Address, Network
        >>> script = PlutusV2Script.load("test/resources/scriptV2.plutus")
        >>> Address(plutus_script_hash(script), network=Network.TESTNET).encode()
        'addr_test1wrmz3pjz4dmfxj0fc0a0eyw69tp6h7mpndzf9g3kttq9cqqqw47ym'
    """

    @property
    def version(self) -> int:
        raise NotImplementedError("")

    def to_shallow_primitive(self) -> bytes:
        return bytes(self)

    @classmethod
    def from_primitive(
        cls: Type[PlutusScript], value: Any, type_args: Optional[tuple] = None
    ) -> PlutusScript:
        if not isinstance(value, (bytes, bytearray)):
            raise DeserializeException(f"Expect bytes, got {type(value)} instead.")
        return cls(value)

    @classmethod
    def from_version(cls, version: int, script_data: bytes) -> "PlutusScript":
        class_name = f"PlutusV{version}Script"
        script_class = globals().get(class_name)

        if script_class is None:
            raise ValueError(f"No Plutus script class found for version {version}")

        return script_class(script_data)

    def get_script_hash_prefix(self) -> bytes:
        raise NotImplementedError("")

    def __repr__(self):
        return f"{self.__class__.__name__}({self.hex()})"


class PlutusV1Script(PlutusScript):
    def get_script_hash_prefix(self) -> bytes:
        return bytes.fromhex("01")

    @property
    def version(self) -> int:
        return 1


class PlutusV2Script(PlutusScript):
    def get_script_hash_prefix(self) -> bytes:
        return bytes.fromhex("02")

    @property
    def version(self) -> int:
        return 2


class PlutusV3Script(PlutusScript):
    def get_script_hash_prefix(self) -> bytes:
        return bytes.fromhex("03")

    @property
    def version(self) -> int:
        return 3


ScriptType = Union[NativeScript, PlutusScript]
"""Script type. A Union type that contains all valid script types."""


def script_hash(script: ScriptType) -> ScriptHash:
    """Calculates the hash of a script, which could be either native script or plutus script.

    Args:
        script (ScriptType): A script.

    Returns:
        ScriptHash: blake2b hash of the script.
    """
    if isinstance(script, NativeScript):
        return script.hash()
    elif isinstance(script, PlutusScript):
        return ScriptHash(
            blake2b(
                script.get_script_hash_prefix() + script,
                SCRIPT_HASH_SIZE,
                encoder=RawEncoder,
            )
        )
    elif type(script) is bytes:
        return ScriptHash(
            blake2b(bytes.fromhex("01") + script, SCRIPT_HASH_SIZE, encoder=RawEncoder)
        )
    else:
        raise TypeError(f"Unexpected script type: {type(script)}")


@dataclass
class Unit(PlutusData):
    """The default "Unit type" with a 0 constructor ID"""

    CONSTR_ID = 0



"""
Pool parameters for stake pool registration certificate.
"""

from __future__ import annotations

import socket
from dataclasses import dataclass, field
from fractions import Fraction
from typing import List, Optional, Type, Union

from pycardano.crypto.bech32 import bech32_decode, decode, encode
from pycardano.exception import DecodingException, DeserializeException
from pycardano.hash import (
    PoolKeyHash,
    PoolMetadataHash,
    RewardAccountHash,
    VerificationKeyHash,
    VrfKeyHash,
)
from pycardano.serialization import (
    ArrayCBORSerializable,
    CBORSerializable,
    OrderedSet,
    limit_primitive_type,
)

__all__ = [
    "PoolId",
    "PoolMetadata",
    "PoolOperator",
    "PoolParams",
    "Relay",
    "SingleHostAddr",
    "SingleHostName",
    "MultiHostName",
    "is_bech32_cardano_pool_id",
]


def is_bech32_cardano_pool_id(pool_id: str) -> bool:
    """Check if a string is a valid Cardano stake pool ID in bech32 format."""
    if pool_id is None or not pool_id.startswith("pool"):
        return False
    return bech32_decode(pool_id) != (None, None, None)


@dataclass(frozen=True)
class PoolId(CBORSerializable):
    value: str

    def __post_init__(self):
        if not is_bech32_cardano_pool_id(self.value):
            raise ValueError(
                "Invalid PoolId format. The PoolId should be a valid Cardano stake pool ID in bech32 format."
            )

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value

    def to_primitive(self) -> str:
        return self.value

    @classmethod
    @limit_primitive_type(str)
    def from_primitive(cls: Type[PoolId], value: str) -> PoolId:
        return cls(value)


@dataclass(repr=False)
class SingleHostAddr(ArrayCBORSerializable):
    _CODE: int = field(init=False, default=0)

    port: Optional[int]
    ipv4: Optional[Union[str, bytes]]
    ipv6: Optional[Union[str, bytes]]

    def __init__(
        self,
        port: Optional[int] = None,
        ipv4: Optional[Union[str, bytes]] = None,
        ipv6: Optional[Union[str, bytes]] = None,
    ):
        super().__init__()

        self._CODE = 0
        self.port = port

        self.ipv4 = self.bytes_to_ipv4(ipv4)
        self.ipv6 = self.bytes_to_ipv6(ipv6)

    @staticmethod
    def ipv4_to_bytes(ip_address: Optional[str | bytes] = None) -> bytes | None:
        """
        Convert IPv4 address to bytes format.
        Args:
            ip_address: The IPv4 address in human-readable format.

        Returns:
            bytes: IPv4 address in bytes format.
        """
        if isinstance(ip_address, str):
            return socket.inet_aton(ip_address)
        elif isinstance(ip_address, bytes):
            return ip_address
        else:
            return None

    @staticmethod
    def ipv6_to_bytes(ip_address: Optional[str | bytes] = None) -> bytes | None:
        """
        Convert IPv6 address to bytes format.
        Args:
            ip_address: The IPv6 address in human-readable format.

        Returns:
            The IPv6 address in bytes format.
        """
        if isinstance(ip_address, str):
            return socket.inet_pton(socket.AF_INET6, ip_address)
        elif isinstance(ip_address, bytes):
            return ip_address
        else:
            return None

    @staticmethod
    def bytes_to_ipv4(bytes_ip_address: Optional[str | bytes] = None) -> str | None:
        """
        Convert IPv4 address in bytes to human-readable format.
        Args:
            bytes_ip_address: The IPv4 address in bytes format.
        Returns:
            The IPv4 address in human-readable format.
        """
        if isinstance(bytes_ip_address, str):
            return bytes_ip_address
        elif isinstance(bytes_ip_address, bytes):
            return socket.inet_ntoa(bytes_ip_address)
        else:
            return None

    @staticmethod
    def bytes_to_ipv6(bytes_ip_address: Optional[str | bytes] = None) -> str | None:
        """
        Convert IPv6 address in bytes to human-readable format.
        Args:
            bytes_ip_address: The IPv6 address in bytes format.
        Returns:
            The IPv6 address in human-readable format.
        """
        if isinstance(bytes_ip_address, str):
            return bytes_ip_address
        elif isinstance(bytes_ip_address, bytes):
            return socket.inet_ntop(socket.AF_INET6, bytes_ip_address)
        else:
            return None

    def to_primitive(self) -> list:
        return [
            self._CODE,
            self.port,
            self.ipv4_to_bytes(self.ipv4),
            self.ipv6_to_bytes(self.ipv6),
        ]

    @classmethod
    @limit_primitive_type(list, tuple)
    def from_primitive(
        cls: Type[SingleHostAddr], values: Union[list, tuple]
    ) -> SingleHostAddr:
        if values[0] == 0:
            return cls(
                port=values[1],
                ipv4=values[2],
                ipv6=values[3],
            )
        else:
            raise DeserializeException(f"Invalid SingleHostAddr type {values[0]}")


@dataclass(repr=False)
class SingleHostName(ArrayCBORSerializable):
    _CODE: int = field(init=False, default=1)

    port: Optional[int]
    dns_name: Optional[str]

    def __post_init__(self):
        self._CODE = 1

    @classmethod
    @limit_primitive_type(list, tuple)
    def from_primitive(
        cls: Type[SingleHostName], values: Union[list, tuple]
    ) -> SingleHostName:
        if values[0] == 1:
            return cls(
                port=values[1],
                dns_name=values[2],
            )
        else:
            raise DeserializeException(f"Invalid SingleHostName type {values[0]}")


@dataclass(repr=False)
class MultiHostName(ArrayCBORSerializable):
    _CODE: int = field(init=False, default=2)

    dns_name: Optional[str]

    def __post_init__(self):
        self._CODE = 2

    @classmethod
    @limit_primitive_type(list, tuple)
    def from_primitive(
        cls: Type[MultiHostName], values: Union[list, tuple]
    ) -> MultiHostName:
        if values[0] == 2:
            return cls(
                dns_name=values[1],
            )
        else:
            raise DeserializeException(f"Invalid MultiHostName type {values[0]}")


Relay = Union[SingleHostAddr, SingleHostName, MultiHostName]


@dataclass(repr=False)
class PoolMetadata(ArrayCBORSerializable):
    url: str
    pool_metadata_hash: PoolMetadataHash


@dataclass(repr=False)
class PoolParams(ArrayCBORSerializable):
    operator: PoolKeyHash
    vrf_keyhash: VrfKeyHash
    pledge: int
    cost: int
    margin: Fraction
    reward_account: RewardAccountHash
    pool_owners: Union[List[VerificationKeyHash], OrderedSet[VerificationKeyHash]]
    relays: Optional[List[Relay]] = None
    pool_metadata: Optional[PoolMetadata] = None
    id: Optional[PoolId] = field(default=None, metadata={"optional": True})


@dataclass(repr=False)
class PoolOperator(CBORSerializable):
    pool_key_hash: PoolKeyHash

    def __init__(self, pool_key_hash: PoolKeyHash):
        self.pool_key_hash = pool_key_hash

    def __repr__(self):
        return f"{self.encode()}"

    def __bytes__(self):
        return self.pool_key_hash.payload

    def encode(self) -> str:
        """Encode the pool key hash in Bech32 format.

        More info about Bech32 `here <https://github.com/bitcoin/bips/blob/master/bip-0173.mediawiki#Bech32>`_.

        Returns:
            str: Encoded pool key hash in Bech32.

        Examples:
            >>> pool_key_hash = PoolKeyHash(bytes.fromhex("cc30497f4ff962f4c1dca54cceefe39f86f1d7179668009f8eb71e59"))
            >>> pool_operator = PoolOperator(pool_key_hash=pool_key_hash)
            >>> print(pool_operator.encode())
            pool1escyjl60l930fswu54xvamlrn7r0r4chje5qp8uwku09j7x68x6
        """
        return encode("pool", self.pool_key_hash.payload)

    @classmethod
    def decode(cls, data: str) -> PoolOperator:
        """Decode a bech32 string into a pool operator object.

        Args:
            data (str): Bech32-encoded string.

        Returns:
            PoolOperator: Decoded pool operator.

        Raises:
            DecodingException: When the input string is not a valid Shelley address.

        Examples:
            >>> pool_operator = PoolOperator.decode("pool1escyjl60l930fswu54xvamlrn7r0r4chje5qp8uwku09j7x68x6")
            >>> pool_key_hash = PoolKeyHash(bytes.fromhex("cc30497f4ff962f4c1dca54cceefe39f86f1d7179668009f8eb71e59"))
            >>> assert pool_operator == PoolOperator(pool_key_hash)
        """
        return cls.from_primitive(data)

    def to_shallow_primitive(self) -> bytes:
        return self.pool_key_hash.to_primitive()

    @classmethod
    @limit_primitive_type(bytes, str)
    def from_primitive(
        cls: Type[PoolOperator], value: Union[bytes, str]
    ) -> PoolOperator:
        # Convert string to bytes
        if isinstance(value, str):
            # Check if bech32 poolid
            if value.startswith("pool"):
                value = bytes(decode(value))
            else:
                try:
                    value = bytes.fromhex(value)
                except Exception as e:
                    raise DecodingException(
                        f"Failed to decode pool id string: {e}"
                    ) from e
        return cls(PoolKeyHash.from_primitive(value))



from typing import Dict, List, Optional, Tuple, Union

import requests
from cachetools import Cache, LRUCache, TTLCache

from pycardano.address import Address
from pycardano.backend.base import ChainContext, GenesisParameters, ProtocolParameters
from pycardano.backend.blockfrost import _try_fix_script
from pycardano.hash import DatumHash, ScriptHash
from pycardano.network import Network
from pycardano.plutus import ExecutionUnits, PlutusScript
from pycardano.serialization import RawCBOR
from pycardano.transaction import (
    Asset,
    AssetName,
    MultiAsset,
    TransactionInput,
    TransactionOutput,
    UTxO,
    Value,
)

__all__ = ["KupoChainContextExtension"]


def extract_asset_info(asset_hash: str) -> Tuple[str, ScriptHash, AssetName]:
    split_result = asset_hash.split(".")

    if len(split_result) == 1:
        policy_hex, asset_name_hex = split_result[0], ""
    elif len(split_result) == 2:
        policy_hex, asset_name_hex = split_result
    else:
        raise ValueError(f"Unable to parse asset hash: {asset_hash}")

    policy = ScriptHash.from_primitive(policy_hex)
    asset_name = AssetName.from_primitive(asset_name_hex)

    return policy_hex, policy, asset_name


class KupoChainContextExtension(ChainContext):
    _wrapped_backend: ChainContext
    _kupo_url: Optional[str]
    _utxo_cache: Cache
    _datum_cache: Cache
    _refetch_chain_tip_interval: int

    def __init__(
        self,
        wrapped_backend: ChainContext,
        kupo_url: Optional[str] = None,
        refetch_chain_tip_interval: int = 10,
        utxo_cache_size: int = 1000,
        datum_cache_size: int = 1000,
    ):
        self._kupo_url = kupo_url
        self._wrapped_backend = wrapped_backend
        self._refetch_chain_tip_interval = refetch_chain_tip_interval
        self._utxo_cache = TTLCache(
            ttl=self._refetch_chain_tip_interval, maxsize=utxo_cache_size
        )
        self._datum_cache = LRUCache(maxsize=datum_cache_size)

    @property
    def genesis_param(self) -> GenesisParameters:
        """Get chain genesis parameters"""

        return self._wrapped_backend.genesis_param

    @property
    def protocol_param(self) -> ProtocolParameters:
        """Get current protocol parameters"""
        return self._wrapped_backend.protocol_param

    @property
    def network(self) -> Network:
        """Get current network"""
        return self._wrapped_backend.network

    @property
    def epoch(self) -> int:
        """Current epoch number"""
        return self._wrapped_backend.epoch

    @property
    def last_block_slot(self) -> int:
        """Last block slot"""
        return self._wrapped_backend.last_block_slot

    def _utxos(self, address: str) -> List[UTxO]:
        """Get all UTxOs associated with an address.

        Args:
            address (str): An address encoded with bech32.

        Returns:
            List[UTxO]: A list of UTxOs.
        """
        key = (self.last_block_slot, address)
        if key in self._utxo_cache:
            return self._utxo_cache[key]

        if self._kupo_url:
            utxos = self._utxos_kupo(address)
        else:
            utxos = self._wrapped_backend.utxos(address)

        self._utxo_cache[key] = utxos

        return utxos

    def _get_datum_from_kupo(self, datum_hash: str) -> Optional[RawCBOR]:
        """Get datum from Kupo.

        Args:
            datum_hash (str): A datum hash.

        Returns:
            Optional[RawCBOR]: A datum.
        """
        datum = self._datum_cache.get(datum_hash, None)

        if datum is not None:
            return datum

        if self._kupo_url is None:
            raise AssertionError(
                "kupo_url object attribute has not been assigned properly."
            )

        kupo_datum_url = self._kupo_url + "/datums/" + datum_hash
        datum_result = requests.get(kupo_datum_url).json()
        if datum_result and datum_result["datum"] != datum_hash:
            datum = RawCBOR(bytes.fromhex(datum_result["datum"]))

        self._datum_cache[datum_hash] = datum
        return datum

    def _utxos_kupo(self, address: str) -> List[UTxO]:
        """Get all UTxOs associated with an address with Kupo.
        Since UTxO querying will be deprecated from Ogmios in next
        major release: https://ogmios.dev/mini-protocols/local-state-query/.

        Args:
            address (str): An address encoded with bech32.

        Returns:
            List[UTxO]: A list of UTxOs.
        """
        if self._kupo_url is None:
            raise AssertionError(
                "kupo_url object attribute has not been assigned properly."
            )

        kupo_utxo_url = self._kupo_url + "/matches/" + address + "?unspent"
        results = requests.get(kupo_utxo_url).json()

        utxos = []

        for result in results:
            tx_id = result["transaction_id"]
            index = result["output_index"]

            if result["spent_at"] is None:
                tx_in = TransactionInput.from_primitive([tx_id, index])

                lovelace_amount = result["value"]["coins"]

                script = None
                script_hash = result.get("script_hash", None)
                if script_hash:
                    kupo_script_url = self._kupo_url + "/scripts/" + script_hash
                    script = requests.get(kupo_script_url).json()
                    ver = int(script["language"].removeprefix("plutus:v"))
                    if 1 <= ver <= 3:
                        script = PlutusScript.from_version(
                            ver, bytes.fromhex(script["script"])
                        )
                        script = _try_fix_script(script_hash, script)
                    else:
                        raise ValueError("Unknown plutus script type")

                datum = None
                datum_hash = (
                    DatumHash.from_primitive(result["datum_hash"])
                    if result["datum_hash"]
                    else None
                )
                if datum_hash and result.get("datum_type", "inline"):
                    datum = self._get_datum_from_kupo(result["datum_hash"])

                if not result["value"]["assets"]:
                    tx_out = TransactionOutput(
                        Address.from_primitive(address),
                        amount=lovelace_amount,
                        datum_hash=datum_hash,
                        datum=datum,
                        script=script,
                    )
                else:
                    multi_assets = MultiAsset()

                    for asset, quantity in result["value"]["assets"].items():
                        policy_hex, policy, asset_name_hex = extract_asset_info(asset)
                        multi_assets.setdefault(policy, Asset())[
                            asset_name_hex
                        ] = quantity

                    tx_out = TransactionOutput(
                        Address.from_primitive(address),
                        amount=Value(lovelace_amount, multi_assets),
                        datum_hash=datum_hash,
                        datum=datum,
                        script=script,
                    )
                utxos.append(UTxO(tx_in, tx_out))
            else:
                continue

        return utxos

    def submit_tx_cbor(self, cbor: Union[bytes, str]):
        """Submit a transaction to the blockchain.

        Args:
            cbor (Union[bytes, str]): The transaction to be submitted.

        Raises:
            :class:`InvalidArgumentException`: When the transaction is invalid.
            :class:`TransactionFailedException`: When fails to submit the transaction to blockchain.
        """
        return self._wrapped_backend.submit_tx_cbor(cbor)

    def evaluate_tx_cbor(self, cbor: Union[bytes, str]) -> Dict[str, ExecutionUnits]:
        """Evaluate execution units of a transaction.

        Args:
            cbor (Union[bytes, str]): The serialized transaction to be evaluated.

        Returns:
            Dict[str, ExecutionUnits]: A list of execution units calculated for each of the transaction's redeemers

        Raises:
            :class:`TransactionFailedException`: When fails to evaluate the transaction.
        """
        return self._wrapped_backend.evaluate_tx_cbor(cbor)



import time
from typing import Dict, List, Optional, Union

from cachetools import Cache, LRUCache, TTLCache, func
from ogmios.client import Client as OgmiosClient
from ogmios.datatypes import Address as OgmiosAddress
from ogmios.datatypes import Era as OgmiosEra
from ogmios.datatypes import ProtocolParameters as OgmiosProtocolParameters
from ogmios.datatypes import Tip as OgmiosTip
from ogmios.datatypes import TxOutputReference as OgmiosTxOutputReference
from ogmios.datatypes import Utxo as OgmiosUtxo
from ogmios.utils import GenesisParameters as OgmiosGenesisParameters
from ogmios.utils import get_current_era

from pycardano.address import Address
from pycardano.backend.base import ChainContext, GenesisParameters, ProtocolParameters
from pycardano.backend.kupo import KupoChainContextExtension
from pycardano.hash import DatumHash, ScriptHash
from pycardano.network import Network
from pycardano.plutus import (
    PLUTUS_V1_COST_MODEL,
    PLUTUS_V2_COST_MODEL,
    ExecutionUnits,
    PlutusScript,
)
from pycardano.serialization import RawCBOR
from pycardano.transaction import (
    Asset,
    AssetName,
    MultiAsset,
    TransactionInput,
    TransactionOutput,
    UTxO,
    Value,
)

ALONZO_COINS_PER_UTXO_WORD = 34482
DEFAULT_REFETCH_INTERVAL = 1000

__all__ = ["OgmiosV6ChainContext", "OgmiosChainContext", "KupoOgmiosV6ChainContext"]


class OgmiosV6ChainContext(ChainContext):
    """Ogmios chain context for use with PyCardano"""

    _network: Network
    _client: OgmiosClient
    _service_name: str
    _last_known_block_slot: int
    _last_chain_tip_fetch: float
    _genesis_param: Optional[GenesisParameters]
    _protocol_param: Optional[OgmiosProtocolParameters]
    _utxo_cache: Cache
    _datum_cache: Cache

    def __init__(
        self,
        host: str = "localhost",
        port: int = 1337,
        path: str = "",
        secure: bool = False,
        refetch_chain_tip_interval: Optional[float] = None,
        utxo_cache_size: int = 10000,
        datum_cache_size: int = 10000,
        network: Network = Network.TESTNET,
        additional_headers: Optional[dict] = None,
    ):
        self.host = host
        self.port = port
        self.path = path
        self.secure = secure
        self.additional_headers = additional_headers or {}
        self._network = network
        self._service_name = "ogmios"
        self._last_known_block_slot = 0
        self._refetch_chain_tip_interval = (
            refetch_chain_tip_interval
            if refetch_chain_tip_interval is not None
            else DEFAULT_REFETCH_INTERVAL
        )
        self._last_chain_tip_fetch = 0
        self._genesis_param = None
        self._protocol_param = None

        self._utxo_cache = TTLCache(
            ttl=self._refetch_chain_tip_interval, maxsize=utxo_cache_size
        )
        self._datum_cache = LRUCache(maxsize=datum_cache_size)

    def _query_current_era(self) -> OgmiosEra:
        with OgmiosClient(
            self.host, self.port, self.path, self.secure, self.additional_headers
        ) as client:
            return get_current_era(client)

    def _query_current_epoch(self) -> int:
        with OgmiosClient(
            self.host, self.port, self.path, self.secure, self.additional_headers
        ) as client:
            epoch, _ = client.query_epoch.execute()
            return epoch

    def _query_chain_tip(self) -> OgmiosTip:
        with OgmiosClient(
            self.host, self.port, self.path, self.secure, self.additional_headers
        ) as client:
            tip, _ = client.query_network_tip.execute()
            return tip

    def _query_utxos_by_address(self, address: Address) -> List[OgmiosUtxo]:
        with OgmiosClient(
            self.host, self.port, self.path, self.secure, self.additional_headers
        ) as client:
            utxos, _ = client.query_utxo.execute([address])
            return utxos

    def _query_utxos_by_tx_id(self, tx_id: str, index: int) -> List[OgmiosUtxo]:
        with OgmiosClient(
            self.host, self.port, self.path, self.secure, self.additional_headers
        ) as client:
            utxos, _ = client.query_utxo.execute(
                [OgmiosTxOutputReference(tx_id, index)]
            )
            return utxos

    def _is_chain_tip_updated(self):
        # fetch at most every twenty seconds!
        if time.time() - self._last_chain_tip_fetch < self._refetch_chain_tip_interval:
            return False
        self._last_chain_tip_fetch = time.time()
        slot = self.last_block_slot
        if self._last_known_block_slot < slot:
            self._last_known_block_slot = slot
            return True
        else:
            return False

    @staticmethod
    def _fraction_parser(fraction: str) -> float:
        x, y = fraction.split("/")
        return int(x) / int(y)

    @property
    def protocol_param(self) -> ProtocolParameters:
        if not self._protocol_param or self._is_chain_tip_updated():
            self._protocol_param = self._fetch_protocol_param()
        return self._protocol_param

    def _fetch_protocol_param(self) -> ProtocolParameters:
        with OgmiosClient(
            self.host, self.port, self.path, self.secure, self.additional_headers
        ) as client:
            protocol_parameters, _ = client.query_protocol_parameters.execute()
            pyc_protocol_params = ProtocolParameters(
                min_fee_constant=protocol_parameters.min_fee_constant.lovelace,
                min_fee_coefficient=protocol_parameters.min_fee_coefficient,
                min_pool_cost=protocol_parameters.min_stake_pool_cost.lovelace,
                max_block_size=protocol_parameters.max_block_body_size.get("bytes"),
                max_tx_size=protocol_parameters.max_transaction_size.get("bytes"),
                max_block_header_size=protocol_parameters.max_block_header_size.get(
                    "bytes"
                ),
                key_deposit=protocol_parameters.stake_credential_deposit.lovelace,
                pool_deposit=protocol_parameters.stake_pool_deposit.lovelace,
                pool_influence=eval(protocol_parameters.stake_pool_pledge_influence),
                monetary_expansion=eval(protocol_parameters.monetary_expansion),
                treasury_expansion=eval(protocol_parameters.treasury_expansion),
                decentralization_param=None,  # type: ignore[arg-type]
                extra_entropy=protocol_parameters.extra_entropy,
                protocol_major_version=protocol_parameters.version.get("major"),
                protocol_minor_version=protocol_parameters.version.get("minor"),
                min_utxo=None,  # type: ignore[arg-type]
                price_mem=eval(
                    protocol_parameters.script_execution_prices.get("memory")
                ),
                price_step=eval(protocol_parameters.script_execution_prices.get("cpu")),
                max_tx_ex_mem=protocol_parameters.max_execution_units_per_transaction.get(
                    "memory"
                ),
                max_tx_ex_steps=protocol_parameters.max_execution_units_per_transaction.get(
                    "cpu"
                ),
                max_block_ex_mem=protocol_parameters.max_execution_units_per_block.get(
                    "memory"
                ),
                max_block_ex_steps=protocol_parameters.max_execution_units_per_block.get(
                    "cpu"
                ),
                max_val_size=protocol_parameters.max_value_size.get("bytes"),
                collateral_percent=protocol_parameters.collateral_percentage,
                max_collateral_inputs=protocol_parameters.max_collateral_inputs,
                coins_per_utxo_word=ALONZO_COINS_PER_UTXO_WORD,
                coins_per_utxo_byte=protocol_parameters.min_utxo_deposit_coefficient,
                cost_models=self._parse_cost_models(
                    protocol_parameters.plutus_cost_models
                ),
                maximum_reference_scripts_size=protocol_parameters.max_ref_script_size,
                min_fee_reference_scripts=protocol_parameters.min_fee_ref_scripts,
            )
            return pyc_protocol_params

    @property
    def genesis_param(self) -> GenesisParameters:
        if not self._genesis_param or self._is_chain_tip_updated():
            # TODO transform to PyCardano GenesisParameters?
            self._genesis_param = self._fetch_genesis_param()  # type: ignore[assignment]

            # Update the refetch interval if we haven't calculated it yet
            if (
                self._refetch_chain_tip_interval == DEFAULT_REFETCH_INTERVAL
                and self._genesis_param is not None
                and self._genesis_param.slot_length is not None
                and self._genesis_param.active_slots_coefficient is not None
            ):
                self._refetch_chain_tip_interval = (
                    self._genesis_param.slot_length
                    / float(self._genesis_param.active_slots_coefficient)
                )
        return self._genesis_param  # type: ignore[return-value]

    def _fetch_genesis_param(self) -> OgmiosGenesisParameters:
        with OgmiosClient(
            self.host, self.port, self.path, self.secure, self.additional_headers
        ) as client:
            return OgmiosGenesisParameters(client, self._query_current_era())

    @property
    def network(self) -> Network:
        return self._network

    @property
    def epoch(self) -> int:
        return self._query_current_epoch()

    @property
    @func.ttl_cache(ttl=1)
    def last_block_slot(self) -> int:
        tip = self._query_chain_tip()
        return tip.slot

    def _utxos(self, address: str) -> List[UTxO]:
        key = (self.last_block_slot, address)
        if key in self._utxo_cache:
            return self._utxo_cache[key]

        utxos = self._utxos_ogmios(OgmiosAddress(address=address))

        self._utxo_cache[key] = utxos

        return utxos

    def _check_utxo_unspent(self, tx_id: str, index: int) -> bool:
        results = self._query_utxos_by_tx_id(tx_id, index)
        return len(results) > 0

    def _utxos_ogmios(self, address: Address) -> List[OgmiosUtxo]:
        """Get all UTxOs associated with an address with Ogmios.

        Args:
            address (str): An address encoded with bech32.

        Returns:
            List[UTxO]: A list of UTxOs.
        """
        results = self._query_utxos_by_address(address)

        utxos = []
        for result in results:
            utxos.append(self._utxo_from_ogmios_result(result))

        return utxos

    def _utxo_from_ogmios_result(self, utxo: OgmiosUtxo) -> UTxO:
        """Convert an Ogmios UTxO result to a PyCardano UTxO."""
        tx_in = TransactionInput.from_primitive([utxo.tx_id, utxo.index])
        lovelace_amount = utxo.value.get("ada").get("lovelace", 0)
        script = utxo.script
        if script:
            # TODO: Need to test with native scripts
            if script["language"].startswith("plutus:v"):
                script = PlutusScript.from_version(
                    int(script["language"].removeprefix("plutus:v")),
                    bytes.fromhex(script["cbor"]),
                )
            else:
                raise ValueError("Unknown plutus script type")
        datum_hash = (
            DatumHash.from_primitive(utxo.datum_hash) if utxo.datum_hash else None
        )
        datum = None
        if utxo.datum and utxo.datum != utxo.datum_hash:
            datum = RawCBOR(bytes.fromhex(utxo.datum))
        if set(utxo.value.keys()) == {"ada"}:
            tx_out = TransactionOutput(
                Address.from_primitive(utxo.address),
                amount=lovelace_amount,
                datum_hash=datum_hash,
                datum=datum,
                script=script,
            )
        else:
            multi_assets = MultiAsset()
            for asset_hex, token in utxo.value.items():
                if asset_hex != "ada":
                    for token_name_hex, quantity in token.items():
                        policy = ScriptHash.from_primitive(asset_hex)
                        token_name = AssetName.from_primitive(token_name_hex)
                        multi_assets.setdefault(policy, Asset())[token_name] = quantity

            tx_out = TransactionOutput(
                Address.from_primitive(utxo.address),
                amount=Value(lovelace_amount, multi_assets),
                datum_hash=datum_hash,
                datum=datum,
                script=script,
            )
        pyc_utxo = UTxO(tx_in, tx_out)
        return pyc_utxo

    def utxo_by_tx_id(self, tx_id: str, index: int) -> Optional[UTxO]:
        utxos = self._query_utxos_by_tx_id(tx_id, index)
        if len(utxos) > 0:
            return self._utxo_from_ogmios_result(utxos[0])
        return None

    def query_account_reward_summaries(
        self, scripts: Optional[List[str]] = None, keys: Optional[List[str]] = None
    ) -> List[dict]:
        with OgmiosClient(
            self.host, self.port, self.path, self.secure, self.additional_headers
        ) as client:
            summaries, _ = client.query_reward_account_summaries.execute(
                scripts=scripts, keys=keys
            )
            return summaries

    def submit_tx_cbor(self, cbor: Union[bytes, str]):
        if isinstance(cbor, bytes):
            cbor = cbor.hex()
        with OgmiosClient(
            self.host, self.port, self.path, self.secure, self.additional_headers
        ) as client:
            client.submit_transaction.execute(cbor)

    def evaluate_tx_cbor(self, cbor: Union[bytes, str]) -> Dict[str, ExecutionUnits]:
        if isinstance(cbor, bytes):
            cbor = cbor.hex()
        with OgmiosClient(
            self.host, self.port, self.path, self.secure, self.additional_headers
        ) as client:
            result, _ = client.evaluate_transaction.execute(cbor)
            result_dict = {}
            for res in result:
                purpose = res["validator"]["purpose"]
                # Hotfix: this purpose has been renamed in the latest version of Ogmios
                if purpose == "withdraw":
                    purpose = "withdrawal"
                result_dict[f"{purpose}:{res['validator']['index']}"] = ExecutionUnits(
                    mem=res["budget"]["memory"],
                    steps=res["budget"]["cpu"],
                )
            return result_dict

    def _parse_cost_models(self, plutus_cost_models):
        ogmios_cost_models = plutus_cost_models or {}

        cost_models = {}
        if "plutus:v1" in ogmios_cost_models:
            cost_models["PlutusV1"] = dict(
                zip(
                    sorted(PLUTUS_V1_COST_MODEL.keys()),
                    ogmios_cost_models["plutus:v1"].copy(),
                )
            )
        if "plutus:v2" in ogmios_cost_models:
            cost_models["PlutusV2"] = dict(
                zip(
                    sorted(PLUTUS_V2_COST_MODEL.keys()),
                    ogmios_cost_models["plutus:v2"].copy(),
                )
            )
        if "plutus:v3" in ogmios_cost_models:
            cost_models["PlutusV3"] = {}
            width = len(f'{len(ogmios_cost_models["plutus:v3"])}')
            for i, v in enumerate(ogmios_cost_models["plutus:v3"].copy()):
                cost_models["PlutusV3"][f"{i:0{width}d}"] = v
        return cost_models


class OgmiosChainContext(OgmiosV6ChainContext):
    """An alias of OgmiosV6ChainContext for backwards compatibility."""

    pass


def KupoOgmiosV6ChainContext(
    host: str,
    port: int,
    path: str,
    secure: bool,
    refetch_chain_tip_interval: Optional[float] = None,
    utxo_cache_size: int = 10000,
    datum_cache_size: int = 10000,
    network: Network = Network.TESTNET,
    additional_headers: Optional[dict] = None,
    kupo_url: Optional[str] = None,
) -> KupoChainContextExtension:
    return KupoChainContextExtension(
        OgmiosV6ChainContext(
            host,
            port,
            path,
            secure,
            refetch_chain_tip_interval,
            utxo_cache_size,
            datum_cache_size,
            network,
            additional_headers,
        ),
        kupo_url,
    )



"""Cardano native script"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import ClassVar, List, Type, Union, cast

from nacl.encoding import RawEncoder
from nacl.hash import blake2b

from pycardano.exception import DeserializeException
from pycardano.hash import SCRIPT_HASH_SIZE, ScriptHash, VerificationKeyHash
from pycardano.serialization import (
    ArrayCBORSerializable,
    Primitive,
    limit_primitive_type,
    list_hook,
)
from pycardano.types import JsonDict

__all__ = [
    "NativeScript",
    "ScriptPubkey",
    "ScriptAll",
    "ScriptAny",
    "ScriptNofK",
    "InvalidBefore",
    "InvalidHereAfter",
]


@dataclass
class NativeScript(ArrayCBORSerializable):
    json_tag: ClassVar[str]
    json_field: ClassVar[str]

    @classmethod
    @limit_primitive_type(list, tuple)
    def from_primitive(
        cls: Type[NativeScript], value: list
    ) -> Union[
        ScriptPubkey, ScriptAll, ScriptAny, ScriptNofK, InvalidBefore, InvalidHereAfter
    ]:
        script_type: int = value[0]
        if script_type == ScriptPubkey._TYPE:
            return super(NativeScript, ScriptPubkey).from_primitive(value[1:])
        elif script_type == ScriptAll._TYPE:
            return super(NativeScript, ScriptAll).from_primitive(value[1:])
        elif script_type == ScriptAny._TYPE:
            return super(NativeScript, ScriptAny).from_primitive(value[1:])
        elif script_type == ScriptNofK._TYPE:
            return super(NativeScript, ScriptNofK).from_primitive(value[1:])
        elif script_type == InvalidBefore._TYPE:
            return super(NativeScript, InvalidBefore).from_primitive(value[1:])
        elif script_type == InvalidHereAfter._TYPE:
            return super(NativeScript, InvalidHereAfter).from_primitive(value[1:])
        else:
            raise DeserializeException(f"Unknown script type indicator: {script_type}")

    def hash(self) -> ScriptHash:
        cbor_bytes = cast(bytes, self.to_cbor())
        return ScriptHash(
            blake2b(bytes(1) + cbor_bytes, SCRIPT_HASH_SIZE, encoder=RawEncoder)
        )

    @classmethod
    def from_dict(
        cls: Type[NativeScript], script_json: JsonDict
    ) -> Union[
        ScriptPubkey, ScriptAll, ScriptAny, ScriptNofK, InvalidBefore, InvalidHereAfter
    ]:
        """Parse a standard native script dictionary (potentially parsed from a JSON file)."""
        script_primitive = cls._script_json_to_primitive(script_json)
        return cls.from_primitive(script_primitive)

    @classmethod
    def _script_json_to_primitive(
        cls: Type[NativeScript], script_json: JsonDict
    ) -> List[Primitive]:
        """Serialize a standard JSON native script into a primitive array"""
        script_type: str = script_json["type"]
        native_script: List[Primitive] = [JSON_TAG_TO_INT[script_type]]

        for key, value in script_json.items():
            if key == "type":
                continue
            elif key == "scripts":
                native_script.append(cls._script_jsons_to_primitive(value))
            else:
                native_script.append(value)
        return native_script

    @classmethod
    def _script_jsons_to_primitive(
        cls: Type[NativeScript], script_jsons: List[JsonDict]
    ) -> List[List[Primitive]]:
        """Parse a list of JSON scripts into a list of primitive arrays"""
        native_script = [cls._script_json_to_primitive(i) for i in script_jsons]
        return native_script

    def to_dict(self) -> JsonDict:
        """Export to standard native script dictionary (potentially to dump to a JSON file)."""
        script: JsonDict = {}
        for value in self.__dict__.values():
            script["type"] = self.json_tag

            if isinstance(value, list):
                script["scripts"] = [i.to_dict() for i in value]
            elif isinstance(value, int):
                script[self.json_field] = value
            else:
                script[self.json_field] = str(value)

        return script


@dataclass
class ScriptPubkey(NativeScript):
    json_tag: ClassVar[str] = "sig"
    json_field: ClassVar[str] = "keyHash"
    _TYPE: int = field(default=0, init=False)

    key_hash: VerificationKeyHash


@dataclass
class ScriptAll(NativeScript):
    json_tag: ClassVar[str] = "all"
    json_field: ClassVar[str] = "scripts"
    _TYPE: int = field(default=1, init=False)

    native_scripts: List[
        Union[
            ScriptPubkey,
            ScriptAll,
            ScriptAny,
            ScriptNofK,
            InvalidBefore,
            InvalidHereAfter,
        ]
    ] = field(metadata={"object_hook": list_hook(NativeScript)})


@dataclass
class ScriptAny(NativeScript):
    json_tag: ClassVar[str] = "any"
    json_field: ClassVar[str] = "scripts"
    _TYPE: int = field(default=2, init=False)

    native_scripts: List[
        Union[
            ScriptPubkey,
            ScriptAll,
            ScriptAny,
            ScriptNofK,
            InvalidBefore,
            InvalidHereAfter,
        ]
    ] = field(metadata={"object_hook": list_hook(NativeScript)})


@dataclass
class ScriptNofK(NativeScript):
    json_tag: ClassVar[str] = "atLeast"
    json_field: ClassVar[str] = "required"
    _TYPE: int = field(default=3, init=False)

    n: int

    native_scripts: List[
        Union[
            ScriptPubkey,
            ScriptAll,
            ScriptAny,
            ScriptNofK,
            InvalidBefore,
            InvalidHereAfter,
        ]
    ] = field(metadata={"object_hook": list_hook(NativeScript)})


@dataclass
class InvalidBefore(NativeScript):
    json_tag: ClassVar[str] = "after"
    json_field: ClassVar[str] = "slot"
    _TYPE: int = field(default=4, init=False)

    before: int


@dataclass
class InvalidHereAfter(NativeScript):
    json_tag: ClassVar[str] = "before"
    json_field: ClassVar[str] = "slot"
    _TYPE: int = field(default=5, init=False)

    after: int


JSON_TAG_TO_INT = {
    ScriptPubkey.json_tag: ScriptPubkey._TYPE,
    ScriptAll.json_tag: ScriptAll._TYPE,
    ScriptAny.json_tag: ScriptAny._TYPE,
    ScriptNofK.json_tag: ScriptNofK._TYPE,
    InvalidBefore.json_tag: InvalidBefore._TYPE,
    InvalidHereAfter.json_tag: InvalidHereAfter._TYPE,
}



# flake8: noqa

from .address import *
from .backend import *
from .certificate import *
from .cip import *
from .coinselection import *
from .crypto import *
from .exception import *
from .governance import *
from .hash import *
from .key import *
from .metadata import *
from .nativescript import *
from .network import *
from .plutus import *
from .pool_params import *
from .serialization import *
from .transaction import *
from .txbuilder import *
from .utils import *
from .witness import *



"""
Cardano CLI Chain Context
"""

import json
import os
import subprocess
import tempfile
import time
from enum import Enum
from functools import partial
from pathlib import Path
from typing import Dict, List, Optional, Union

import docker
from cachetools import Cache, LRUCache, TTLCache, func
from docker.errors import APIError

from pycardano.address import Address
from pycardano.backend.base import (
    ALONZO_COINS_PER_UTXO_WORD,
    ChainContext,
    GenesisParameters,
    ProtocolParameters,
)
from pycardano.cbor import cbor2
from pycardano.exception import (
    CardanoCliError,
    PyCardanoException,
    TransactionFailedException,
)
from pycardano.hash import DatumHash, ScriptHash
from pycardano.nativescript import NativeScript
from pycardano.network import Network
from pycardano.plutus import Datum, PlutusV1Script, PlutusV2Script, RawPlutusData
from pycardano.serialization import RawCBOR
from pycardano.transaction import (
    Asset,
    AssetName,
    MultiAsset,
    TransactionInput,
    TransactionOutput,
    UTxO,
    Value,
)
from pycardano.types import JsonDict
from pycardano.utils import greater_than_version

if greater_than_version((3, 13)):
    from enum import member  # type: ignore[attr-defined]


__all__ = ["CardanoCliChainContext", "CardanoCliNetwork", "DockerConfig"]


def network_magic(magic_number: int) -> List[str]:
    """
    Returns the network magic number for the cardano-cli
    Args:
        magic_number: The network magic number

    Returns:
        The network magic number arguments
    """
    return ["--testnet-magic", str(magic_number)]


class CardanoCliNetwork(Enum):
    """
    Enum class for Cardano Era
    """

    MAINNET = ["--mainnet"]
    TESTNET = ["--testnet-magic", str(1097911063)]
    PREVIEW = ["--testnet-magic", str(2)]
    PREPROD = ["--testnet-magic", str(1)]
    GUILDNET = ["--testnet-magic", str(141)]
    CUSTOM = (
        member(partial(network_magic))
        if greater_than_version((3, 13))
        else partial(network_magic)
    )


class DockerConfig:
    """
    Docker configuration to use the cardano-cli in a Docker container
    """

    container_name: str
    """ The name of the Docker container containing the cardano-cli"""

    host_socket: Optional[Path]
    """ The path to the Docker host socket file"""

    def __init__(self, container_name: str, host_socket: Optional[Path] = None):
        self.container_name = container_name
        self.host_socket = host_socket


class CardanoCliChainContext(ChainContext):
    _binary: Path
    _socket: Optional[Path]
    _config_file: Path
    _network: CardanoCliNetwork
    _last_known_block_slot: int
    _last_chain_tip_fetch: float
    _genesis_param: Optional[GenesisParameters]
    _protocol_param: Optional[ProtocolParameters]
    _utxo_cache: Cache
    _datum_cache: Cache
    _docker_config: Optional[DockerConfig]
    _network_magic_number: Optional[int]

    def __init__(
        self,
        binary: Path,
        socket: Path,
        config_file: Path,
        network: CardanoCliNetwork,
        refetch_chain_tip_interval: Optional[float] = None,
        utxo_cache_size: int = 10000,
        datum_cache_size: int = 10000,
        docker_config: Optional[DockerConfig] = None,
        network_magic_number: Optional[int] = None,
    ):
        if docker_config is None:
            if not binary.exists() or not binary.is_file():
                raise CardanoCliError(f"cardano-cli binary file not found: {binary}")

            # Check the socket path file and set the CARDANO_NODE_SOCKET_PATH environment variable
            try:
                if not socket.exists():
                    raise CardanoCliError(f"cardano-node socket not found: {socket}")
                elif not socket.is_socket():
                    raise CardanoCliError(f"{socket} is not a socket file")

                self._socket = socket
                os.environ["CARDANO_NODE_SOCKET_PATH"] = self._socket.as_posix()
            except CardanoCliError:
                self._socket = None

        self._binary = binary
        self._network = network
        self._config_file = config_file
        self._last_known_block_slot = 0
        self._refetch_chain_tip_interval = (
            refetch_chain_tip_interval
            if refetch_chain_tip_interval is not None
            else 1000
        )
        self._last_chain_tip_fetch = 0
        self._genesis_param = None
        self._protocol_param = None
        if refetch_chain_tip_interval is None:
            self._refetch_chain_tip_interval = float(
                self.genesis_param.slot_length
                / self.genesis_param.active_slots_coefficient
            )

        self._utxo_cache = TTLCache(
            ttl=self._refetch_chain_tip_interval, maxsize=utxo_cache_size
        )
        self._datum_cache = LRUCache(maxsize=datum_cache_size)
        self._docker_config = docker_config
        self._network_magic_number = network_magic_number

    @property
    def _network_args(self) -> List[str]:
        if self._network is CardanoCliNetwork.CUSTOM:
            return self._network.value(self._network_magic_number)
        else:
            return self._network.value

    def _run_command(self, cmd: List[str]) -> str:
        """
        Runs the command in the cardano-cli. If the docker configuration is set, it will run the command in the
        docker container.

        :param cmd: Command as a list of strings
        :return: The stdout if the command runs successfully
        """
        try:
            if self._docker_config:
                docker_config = self._docker_config
                if docker_config.host_socket is None:
                    client = docker.from_env()
                else:
                    client = docker.DockerClient(
                        base_url=docker_config.host_socket.as_posix()
                    )

                container = client.containers.get(docker_config.container_name)

                exec_result = container.exec_run(
                    [self._binary.as_posix()] + cmd, stdout=True, stderr=True
                )

                if exec_result.exit_code == 0:
                    output = exec_result.output.decode()
                    return output
                else:
                    error = exec_result.output.decode()
                    raise CardanoCliError(error)
            else:
                result = subprocess.run(
                    [self._binary.as_posix()] + cmd, capture_output=True, check=True
                )
                return result.stdout.decode().strip()
        except subprocess.CalledProcessError as err:
            raise CardanoCliError(err.stderr.decode()) from err
        except APIError as err:
            raise CardanoCliError(err) from err

    def _query_chain_tip(self) -> JsonDict:
        result = self._run_command(["query", "tip"] + self._network_args)
        return json.loads(result)

    def _query_current_protocol_params(self) -> JsonDict:
        result = self._run_command(
            ["query", "protocol-parameters"] + self._network_args
        )
        return json.loads(result)

    def _query_genesis_config(self) -> JsonDict:
        if not self._config_file.exists() or not self._config_file.is_file():
            raise CardanoCliError(f"Cardano config file not found: {self._config_file}")
        with open(self._config_file, encoding="utf-8") as config_file:
            config_json = json.load(config_file)
            shelly_genesis_file = (
                self._config_file.parent / config_json["ShelleyGenesisFile"]
            )
        if not shelly_genesis_file.exists() or not shelly_genesis_file.is_file():
            raise CardanoCliError(
                f"Shelly Genesis file not found: {shelly_genesis_file}"
            )
        with open(shelly_genesis_file, encoding="utf-8") as genesis_file:
            genesis_json = json.load(genesis_file)
        return genesis_json

    def _get_min_utxo(self) -> int:
        params = self._query_current_protocol_params()
        if "minUTxOValue" in params and params["minUTxOValue"] is not None:
            return params["minUTxOValue"]
        elif (
            "lovelacePerUTxOWord" in params
            and params["lovelacePerUTxOWord"] is not None
        ):
            return params["lovelacePerUTxOWord"]
        else:
            return 0

    @staticmethod
    def _parse_cost_models(cli_result: JsonDict) -> Dict[str, Dict[str, int]]:
        cli_cost_models = cli_result.get("costModels", {})

        cost_models = {}
        if "PlutusScriptV1" in cli_cost_models:
            cost_models["PlutusScriptV1"] = cli_cost_models["PlutusScriptV1"].copy()
        elif "PlutusV1" in cli_cost_models:
            cost_models["PlutusV1"] = cli_cost_models["PlutusV1"].copy()

        if "PlutusScriptV2" in cli_cost_models:
            cost_models["PlutusScriptV2"] = cli_cost_models["PlutusScriptV2"].copy()
        elif "PlutusV2" in cli_cost_models:
            cost_models["PlutusV2"] = cli_cost_models["PlutusV2"].copy()

        # After 8.x.x, cardano-cli returns cost models as a list
        for m in cost_models:
            if isinstance(cost_models[m], list):
                cost_models[m] = {i: v for i, v in enumerate(cost_models[m])}

        return cost_models

    def _is_chain_tip_updated(self):
        # fetch at almost every twenty seconds!
        if time.time() - self._last_chain_tip_fetch < self._refetch_chain_tip_interval:
            return False
        self._last_chain_tip_fetch = time.time()
        result = self._query_chain_tip()
        return float(result["syncProgress"]) != 100.0

    def _fetch_protocol_param(self) -> ProtocolParameters:
        result = self._query_current_protocol_params()
        return ProtocolParameters(
            min_fee_constant=(
                result["minFeeConstant"]
                if "minFeeConstant" in result
                else result["txFeeFixed"]
            ),
            min_fee_coefficient=(
                result["minFeeCoefficient"]
                if "minFeeCoefficient" in result
                else result["txFeePerByte"]
            ),
            max_block_size=result["maxBlockBodySize"],
            max_tx_size=result["maxTxSize"],
            max_block_header_size=result["maxBlockHeaderSize"],
            key_deposit=result["stakeAddressDeposit"],
            pool_deposit=result["stakePoolDeposit"],
            pool_influence=result["poolPledgeInfluence"],
            monetary_expansion=result["monetaryExpansion"],
            treasury_expansion=result["treasuryCut"],
            decentralization_param=result.get("decentralization", 0),
            extra_entropy=result.get("extraPraosEntropy", ""),
            protocol_major_version=result["protocolVersion"]["major"],
            protocol_minor_version=result["protocolVersion"]["minor"],
            min_utxo=self._get_min_utxo(),
            min_pool_cost=result["minPoolCost"],
            price_mem=(
                result["executionUnitPrices"]["priceMemory"]
                if "executionUnitPrices" in result
                else result["executionPrices"]["priceMemory"]
            ),
            price_step=(
                result["executionUnitPrices"]["priceSteps"]
                if "executionUnitPrices" in result
                else result["executionPrices"]["priceSteps"]
            ),
            max_tx_ex_mem=result["maxTxExecutionUnits"]["memory"],
            max_tx_ex_steps=result["maxTxExecutionUnits"]["steps"],
            max_block_ex_mem=result["maxBlockExecutionUnits"]["memory"],
            max_block_ex_steps=result["maxBlockExecutionUnits"]["steps"],
            max_val_size=result["maxValueSize"],
            collateral_percent=result["collateralPercentage"],
            max_collateral_inputs=result["maxCollateralInputs"],
            coins_per_utxo_word=result.get(
                "coinsPerUtxoWord", ALONZO_COINS_PER_UTXO_WORD
            ),
            coins_per_utxo_byte=(
                result["coinsPerUtxoByte"]
                if "coinsPerUtxoByte" in result
                else result.get("utxoCostPerByte", 0) or 0
            ),
            cost_models=self._parse_cost_models(result),
        )

    @property
    def protocol_param(self) -> ProtocolParameters:
        """Get current protocol parameters"""
        if not self._protocol_param or self._is_chain_tip_updated():
            self._protocol_param = self._fetch_protocol_param()
        return self._protocol_param

    @property
    def genesis_param(self) -> GenesisParameters:
        """Get chain genesis parameters"""
        genesis_params = self._query_genesis_config()
        return GenesisParameters(
            active_slots_coefficient=genesis_params["activeSlotsCoeff"],
            update_quorum=genesis_params["updateQuorum"],
            max_lovelace_supply=genesis_params["maxLovelaceSupply"],
            network_magic=genesis_params["networkMagic"],
            epoch_length=genesis_params["epochLength"],
            system_start=genesis_params["systemStart"],
            slots_per_kes_period=genesis_params["slotsPerKESPeriod"],
            slot_length=genesis_params["slotLength"],
            max_kes_evolutions=genesis_params["maxKESEvolutions"],
            security_param=genesis_params["securityParam"],
        )

    @property
    def network(self) -> Network:
        """Cet current network"""
        if self._network == CardanoCliNetwork.MAINNET:
            return Network.MAINNET
        return Network.TESTNET

    @property
    def epoch(self) -> int:
        """Current epoch number"""
        result = self._query_chain_tip()
        return result["epoch"]

    @property
    def era(self) -> int:
        """Current Cardano era"""
        result = self._query_chain_tip()
        return result["era"]

    @property
    @func.ttl_cache(ttl=1)
    def last_block_slot(self) -> int:
        result = self._query_chain_tip()
        return result["slot"]

    def version(self):
        """
        Gets the cardano-cli version
        """
        return self._run_command(["version"])

    @staticmethod
    def _get_script(
        reference_script: dict,
    ) -> Union[PlutusV1Script, PlutusV2Script, NativeScript]:
        """
        Get a script object from a reference script dictionary.
        Args:
            reference_script:

        Returns:

        """
        script_type = reference_script["script"]["type"]
        script_json: JsonDict = reference_script["script"]
        if script_type == "PlutusScriptV1":
            v1script = PlutusV1Script(
                cbor2.loads(bytes.fromhex(script_json["cborHex"]))
            )
            return v1script
        elif script_type == "PlutusScriptV2":
            v2script = PlutusV2Script(
                cbor2.loads(bytes.fromhex(script_json["cborHex"]))
            )
            return v2script
        else:
            return NativeScript.from_dict(script_json)

    def _utxos(self, address: str) -> List[UTxO]:
        """Get all UTxOs associated with an address.

        Args:
            address (str): An address encoded with bech32.

        Returns:
            List[UTxO]: A list of UTxOs.
        """
        key = (self.last_block_slot, address)
        if key in self._utxo_cache:
            return self._utxo_cache[key]

        result = self._run_command(
            ["query", "utxo", "--address", address, "--out-file", "/dev/stdout"]
            + self._network_args
        )

        raw_utxos = json.loads(result)

        utxos = []
        for tx_hash in raw_utxos.keys():
            tx_id, tx_idx = tx_hash.split("#")
            utxo = raw_utxos[tx_hash]
            tx_in = TransactionInput.from_primitive([tx_id, int(tx_idx)])

            value = Value()
            multi_asset = MultiAsset()
            for asset in utxo["value"].keys():
                if asset == "lovelace":
                    value.coin = utxo["value"][asset]
                else:
                    policy_id = asset
                    policy = ScriptHash.from_primitive(policy_id)

                    for asset_hex_name in utxo["value"][asset].keys():
                        asset_name = AssetName.from_primitive(asset_hex_name)
                        amount = utxo["value"][asset][asset_hex_name]
                        multi_asset.setdefault(policy, Asset())[asset_name] = amount

            value.multi_asset = multi_asset

            datum_hash = (
                DatumHash.from_primitive(utxo["datumhash"])
                if utxo.get("datumhash") is not None
                else None
            )

            datum: Optional[Datum] = None

            if utxo.get("datum"):
                datum = RawCBOR(bytes.fromhex(utxo["datum"]))
            elif utxo.get("inlineDatumhash"):
                datum = RawPlutusData.from_dict(utxo["inlineDatum"])

            script = None

            if utxo.get("referenceScript"):
                script = self._get_script(utxo["referenceScript"])

            tx_out = TransactionOutput(
                Address.from_primitive(utxo["address"]),
                amount=value,
                datum_hash=datum_hash,
                datum=datum,
                script=script,
            )

            utxos.append(UTxO(tx_in, tx_out))

        self._utxo_cache[key] = utxos

        return utxos

    def submit_tx_cbor(self, cbor: Union[bytes, str]) -> str:
        """Submit a transaction to the blockchain.

        Args:
            cbor (Union[bytes, str]): The transaction to be submitted.

        Returns:
            str: The transaction hash.

        Raises:
            :class:`TransactionFailedException`: When fails to submit the transaction to blockchain.
            :class:`PyCardanoException`: When fails to retrieve the transaction hash.
        """
        if isinstance(cbor, bytes):
            cbor = cbor.hex()

        with tempfile.NamedTemporaryFile(mode="w") as tmp_tx_file:
            tx_json = {
                "type": f"Witnessed Tx {self.era}Era",
                "description": "Generated by PyCardano",
                "cborHex": cbor,
            }

            tmp_tx_file.write(json.dumps(tx_json))

            tmp_tx_file.flush()

            try:
                self._run_command(
                    [
                        "latest",
                        "transaction",
                        "submit",
                        "--tx-file",
                        tmp_tx_file.name,
                    ]
                    + self._network_args
                )
            except CardanoCliError:
                try:
                    self._run_command(
                        ["transaction", "submit", "--tx-file", tmp_tx_file.name]
                        + self._network_args
                    )
                except CardanoCliError as err:
                    raise TransactionFailedException(
                        "Failed to submit transaction"
                    ) from err

            # Get the transaction ID
            try:
                txid = self._run_command(
                    ["latest", "transaction", "txid", "--tx-file", tmp_tx_file.name]
                )
            except CardanoCliError:
                try:
                    txid = self._run_command(
                        ["transaction", "txid", "--tx-file", tmp_tx_file.name]
                    )
                except CardanoCliError as err:
                    raise PyCardanoException(
                        f"Unable to get transaction id for {tmp_tx_file.name}"
                    ) from err

        return txid



"""
This module contains algorithms that select UTxOs from a parent list to satisfy some output constraints.
"""

import random
from copy import deepcopy
from typing import Iterable, List, Optional, Tuple

from pycardano.address import Address
from pycardano.backend.base import ChainContext
from pycardano.exception import (
    InputUTxODepletedException,
    InsufficientUTxOBalanceException,
    MaxInputCountExceededException,
    UTxOSelectionException,
)
from pycardano.transaction import TransactionOutput, UTxO, Value
from pycardano.utils import max_tx_fee, min_lovelace_post_alonzo

__all__ = ["UTxOSelector", "LargestFirstSelector", "RandomImproveMultiAsset"]

_FAKE_ADDR = Address.from_primitive(
    "addr1q8m9x2zsux7va6w892g38tvchnzahvcd9tykqf3ygnmwta8k2v59pcduem5uw253zwke30x9mwes62kfvqnzg38kuh6q966kg7"
)


class UTxOSelector:
    """UTxOSelector defines an interface through which a subset of UTxOs should be selected from a parent set
    with a selection strategy and given constraints.
    """

    def select(
        self,
        utxos: List[UTxO],
        outputs: List[TransactionOutput],
        context: ChainContext,
        max_input_count: Optional[int] = None,
        include_max_fee: Optional[bool] = True,
        respect_min_utxo: Optional[bool] = True,
        existing_amount: Optional[Value] = None,
    ) -> Tuple[List[UTxO], Value]:
        """From an input list of UTxOs, select a subset of UTxOs whose sum (including ADA and multi-assets)
        is equal to or larger than the sum of a set of outputs.

        Args:
            utxos (List[UTxO]): A list of UTxO to select from.
            outputs (List[TransactionOutput]): A list of transaction outputs which the selected set should satisfy.
            context (ChainContext): A chain context where protocol parameters could be retrieved.
            max_input_count (int): Max number of input UTxOs to select.
            include_max_fee (bool): Have selected UTxOs to cover transaction fee. Defaults to True. If disabled,
                there is a possibility that selected UTxO are not able to cover the fee of the transaction.
            respect_min_utxo (bool): Respect minimum amount of ADA required to hold a multi-asset bundle in the change.
                Defaults to True. If disabled, the selection will not add addition amount of ADA to change even
                when the amount is too small to hold a multi-asset bundle.
            existing_amount (Value): A starting amount already existed before selection. Defaults to 0.

        Returns:
            Tuple[List[UTxO], Value]: A tuple containing:

                selected (List[UTxO]): A list of selected UTxOs.

                changes (Value): Change amount to be returned.

        Raises:
            InsufficientUTxOBalanceException: When total value of input UTxO is less than requested outputs.
            MaxInputCountExceededException: When number of selected UTxOs exceeds `max_input_count`.
            InputUTxODepletedException: When the algorithm has depleted input UTxOs but selection should continue.
            UTxOSelectionException: When selection fails for reasons besides the three above.
        """
        raise NotImplementedError()


class LargestFirstSelector(UTxOSelector):
    """
    Largest first selection algorithm as specified in
    https://github.com/cardano-foundation/CIPs/tree/master/CIP-0002#largest-first.

    This implementation adds transaction fee into consideration.
    """

    def select(
        self,
        utxos: List[UTxO],
        outputs: List[TransactionOutput],
        context: ChainContext,
        max_input_count: Optional[int] = None,
        include_max_fee: Optional[bool] = True,
        respect_min_utxo: Optional[bool] = True,
        existing_amount: Optional[Value] = None,
    ) -> Tuple[List[UTxO], Value]:
        available: List[UTxO] = sorted(utxos, key=lambda utxo: utxo.output.lovelace)
        max_fee = max_tx_fee(context) if include_max_fee else 0
        total_requested = Value(max_fee)
        for o in outputs:
            total_requested += o.amount

        selected = []
        selected_amount = existing_amount if existing_amount is not None else Value()

        while not total_requested <= selected_amount:
            if not available:
                raise InsufficientUTxOBalanceException("UTxO Balance insufficient!")
            to_add = available.pop()
            selected.append(to_add)
            selected_amount += to_add.output.amount
            if max_input_count and len(selected) > max_input_count:
                raise MaxInputCountExceededException(
                    f"Max input count: {max_input_count} exceeded!"
                )

        if respect_min_utxo:
            change = selected_amount - total_requested
            min_change_amount = min_lovelace_post_alonzo(
                TransactionOutput(_FAKE_ADDR, deepcopy(change)), context
            )
            if change.coin < min_change_amount:
                additional, _ = self.select(
                    available,
                    [
                        TransactionOutput(
                            _FAKE_ADDR, Value(min_change_amount - change.coin)
                        )
                    ],
                    context,
                    max_input_count - len(selected) if max_input_count else None,
                    include_max_fee=False,
                    respect_min_utxo=False,
                )
                for u in additional:
                    selected.append(u)
                    selected_amount += u.output.amount
        return selected, selected_amount - total_requested


class RandomImproveMultiAsset(UTxOSelector):
    """Random-improve selection algorithm as specified in
    https://github.com/cardano-foundation/CIPs/tree/master/CIP-0002#random-improve.

    Because the original algorithm does not take multi-assets into consideration, this implementation is slightly
    different from the algorithm. The main modification is that it merges all requested transaction outputs into one,
    including all native assets, and then treat each merged native asset as an individual transaction output request.

    This idea is inspired by Nami wallet: https://github.com/Berry-Pool/nami-wallet/blob/main/src/lib/coinSelection.js

    .. Note::
        Although this implementation is similar to the original Random-improve algorithm, and it is being used by some
        wallets, there are no substantial evidences or proofs showing that this implementation will still be able to
        correctly optimize UTxO selection based on
        `three heuristics <https://github.com/cardano-foundation/CIPs/tree/master/CIP-0002#motivating-principles>`_
        mentioned in the doc.
    """

    def __init__(self, random_generator: Optional[Iterable[int]] = None):
        self.random_generator = iter(random_generator) if random_generator else None

    def _get_next_random(self, utxos: List[UTxO]) -> Tuple[int, UTxO]:
        if not utxos:
            raise InputUTxODepletedException("Input UTxOs depleted!")
        if self.random_generator:
            i = next(self.random_generator, None)
            if i is None:
                raise UTxOSelectionException("Random generator depleted!")
            elif i > len(utxos):
                raise UTxOSelectionException(f"Random index: {i} out of range!")
        else:
            i = random.randint(0, len(utxos) - 1)
        return i, utxos[i]

    def _random_select_subset(
        self,
        amount: Value,
        remaining: List[UTxO],
        selected: List[UTxO],
        selected_amount: Value,
    ):
        while not amount <= selected_amount:
            if not remaining:
                raise InputUTxODepletedException("Input UTxOs depleted!")
            i, to_add = self._get_next_random(remaining)
            selected.append(to_add)
            selected_amount += to_add.output.amount
            remaining.pop(i)

    @staticmethod
    def _split_by_asset(value: Value) -> List[Value]:
        # Extract ADA
        assets = [Value(value.coin)] if value.coin else []

        # Extract native assets
        for policy_id, d in value.multi_asset.items():
            for asset_name, amount in d.items():
                if not amount:
                    # skip 0 amounts
                    continue
                assets.append(
                    Value.from_primitive(
                        [
                            0,
                            {policy_id.payload: {asset_name.payload: amount}},
                        ]
                    )
                )

        return assets

    @staticmethod
    def _get_single_asset_val(value: Value) -> int:
        if value.coin:
            return value.coin
        else:
            return list(list(value.multi_asset.values())[0].values())[0]

    @staticmethod
    def _find_diff_by_former(a: Value, b: Value) -> int:
        """The first argument contains only one asset. Find the absolute difference between this asset and
        the corresponding value of the same asset in the second argument"""
        if a.coin:
            return a.coin - b.coin
        else:
            policy_id = list(a.multi_asset.keys())[0]
            asset_name = list(a.multi_asset[policy_id].keys())[0]
            return a.multi_asset[policy_id].get(asset_name, 0) - b.multi_asset[
                policy_id
            ].get(asset_name, 0)

    def _improve(
        self,
        selected: List[UTxO],
        selected_amount: Value,
        remaining: List[UTxO],
        ideal: Value,
        upper_bound: Value,
        max_input_count: Optional[int] = None,
    ):
        if not remaining or self._find_diff_by_former(ideal, selected_amount) <= 0:
            # In case where there is no remaining UTxOs or we already selected more than ideal,
            # we cannot improve by randomly adding more UTxOs, therefore return immediate.
            return
        if max_input_count is not None and len(selected) > max_input_count:
            raise MaxInputCountExceededException(
                f"Max input count: {max_input_count} exceeded!"
            )

        i, to_add = self._get_next_random(remaining)
        if (
            abs(
                self._find_diff_by_former(ideal, selected_amount + to_add.output.amount)
            )
            < abs(self._find_diff_by_former(ideal, selected_amount))
            and self._find_diff_by_former(
                upper_bound, selected_amount + to_add.output.amount
            )
            >= 0
        ):
            selected.append(to_add)
            selected_amount += to_add.output.amount

        self._improve(
            selected,
            selected_amount,
            remaining[:i] + remaining[i + 1 :],
            ideal,
            upper_bound,
            max_input_count,
        )

    def select(
        self,
        utxos: List[UTxO],
        outputs: List[TransactionOutput],
        context: ChainContext,
        max_input_count: Optional[int] = None,
        include_max_fee: Optional[bool] = True,
        respect_min_utxo: Optional[bool] = True,
        existing_amount: Optional[Value] = None,
    ) -> Tuple[List[UTxO], Value]:
        # Shallow copy the list
        remaining = list(utxos)
        max_fee = max_tx_fee(context) if include_max_fee else 0
        request_sum = Value(max_fee)
        for o in outputs:
            request_sum += o.amount

        assets = self._split_by_asset(request_sum)

        request_sorted = sorted(assets, key=self._get_single_asset_val, reverse=True)

        # Phase 1 - random select
        selected: List[UTxO] = []
        selected_amount = existing_amount if existing_amount is not None else Value()

        for r in request_sorted:
            self._random_select_subset(r, remaining, selected, selected_amount)
            if max_input_count and len(selected) > max_input_count:
                raise MaxInputCountExceededException(
                    f"Max input count: {max_input_count} exceeded!"
                )

        # Phase 2 - improve current selection
        for request in reversed(request_sorted):
            ideal = request + request
            upper_bound = ideal + request
            num_selected_before = len(selected)
            try:
                self._improve(
                    selected,
                    selected_amount,
                    list(remaining),
                    ideal,
                    upper_bound,
                    max_input_count,
                )
            except UTxOSelectionException:
                pass
            new_selected = selected[num_selected_before:]
            remaining = [utxo for utxo in remaining if utxo not in new_selected]

        if respect_min_utxo:
            change = selected_amount - request_sum
            min_change_amount = min_lovelace_post_alonzo(
                TransactionOutput(_FAKE_ADDR, deepcopy(change)), context
            )

            if change.coin < min_change_amount:
                additional, _ = self.select(
                    remaining,
                    [
                        TransactionOutput(
                            _FAKE_ADDR, Value(min_change_amount - change.coin)
                        )
                    ],
                    context,
                    max_input_count - len(selected) if max_input_count else None,
                    include_max_fee=False,
                    respect_min_utxo=False,
                )
                for u in additional:
                    selected.append(u)
                    selected_amount += u.output.amount

        return selected, selected_amount - request_sum



from typing import Optional, Union

from cbor2 import CBORTag, dumps
from cose.algorithms import EdDSA
from cose.headers import KID, Algorithm
from cose.keys import CoseKey
from cose.keys.curves import Ed25519
from cose.keys.keyops import SignOp, VerifyOp
from cose.keys.keyparam import KpAlg, KpKeyOps, KpKty, OKPKpCurve, OKPKpD, OKPKpX
from cose.keys.keytype import KtyOKP
from cose.messages import CoseMessage, Sign1Message

from pycardano.address import Address
from pycardano.crypto import BIP32ED25519PublicKey
from pycardano.key import (
    ExtendedSigningKey,
    ExtendedVerificationKey,
    PaymentVerificationKey,
    SigningKey,
    StakeExtendedSigningKey,
    StakeSigningKey,
    StakeVerificationKey,
)
from pycardano.network import Network

__all__ = ["sign", "verify"]


def sign(
    message: str,
    signing_key: Union[ExtendedSigningKey, SigningKey],
    attach_cose_key: bool = False,
    network: Network = Network.MAINNET,
) -> Union[str, dict]:
    """Sign an arbitrary message with a payment or stake key following CIP-0008.
    Note that a stake key passed in must be of type StakeSigningKey or
    StakeExtendedSigningKey to be detected.

    Args:
        message (str): Message to be signed
        signing_key (SigningKey): Key which is used to sign the message
        attach_cose_key (bool): Whether or not to attach the Cose key to the output
        network (Network): Network to use for the address generation

    Returns:
        Union[str, dict]: A hex-encoded string containing the signed message and verification key.
        In the case of attach_cose_key=True, a dict containing the signed message and Cose key.
    """

    # derive the verification key
    verification_key = signing_key.to_verification_key()
    if isinstance(verification_key, ExtendedVerificationKey):
        verification_key = verification_key.to_non_extended()

    if isinstance(signing_key, StakeSigningKey) or isinstance(
        signing_key, StakeExtendedSigningKey
    ):
        address = Address(
            payment_part=None, staking_part=verification_key.hash(), network=network
        )
    else:
        address = Address(
            payment_part=verification_key.hash(), staking_part=None, network=network
        )

    # create the message object, attach verification key to the header
    msg = Sign1Message(
        phdr={
            Algorithm: EdDSA,
            "address": address.to_primitive(),
        },
        payload=message.encode("utf-8"),
    )

    msg.uhdr = {"hashed": False}

    if not attach_cose_key:
        msg.phdr[KID] = verification_key.to_primitive()

    # build the CoseSign1 signing key from a dictionary
    cose_key = {
        KpKty: KtyOKP,
        OKPKpCurve: Ed25519,
        KpKeyOps: [SignOp, VerifyOp],
        OKPKpD: signing_key.payload,  # private key
        OKPKpX: verification_key.payload,  # public key
    }

    cose_key = CoseKey.from_dict(cose_key)

    msg.key = cose_key  # attach the key to the message

    _message = [
        msg.phdr_encoded,
        msg.uhdr_encoded,
        msg.payload,
        signing_key.sign(msg._sig_structure),
    ]

    encoded = dumps(CBORTag(msg.cbor_tag, _message), default=msg._custom_cbor_encoder)

    # turn the enocded message into a hex string and remove the first byte
    # which is always "d2"
    signed_message = encoded.hex()[2:]

    if attach_cose_key:
        key_to_return = {
            KpKty: KtyOKP,
            KpAlg: EdDSA,
            OKPKpCurve: Ed25519,
            OKPKpX: verification_key.payload,  # public key
        }

        signed_message = {  # type: ignore
            "signature": signed_message,
            "key": CoseKey.from_dict(key_to_return).encode().hex(),
        }

    return signed_message


def verify(
    signed_message: Union[str, dict],
    attach_cose_key: Optional[bool] = None,
) -> dict:
    """Verify the signature of a COSESign1 message and decode its contents following CIP-0008.
    Supports messages signed by browser wallets or `Message.sign()`.

    Args:
        signed_message (Union[str, dict]): Message to be verified
        attach_cose_key (Optional[bool]): Whether or not the Cose key is included with the signed_message.
            This method will try to determine this automatically if not specified.
            Usually if `signed_message` is a dict, this should be true.

    Returns:
        dict: A dict containing whether or not the message is verified, the message contents,
        and an Address which was used to sign the message.
    """

    if attach_cose_key is None:
        # try to determine automatically if the verification key is included in the header
        if isinstance(signed_message, dict):
            attach_cose_key = True
        else:
            attach_cose_key = False

    if attach_cose_key:
        # The cose key is attached as a dict object which contains the verification key
        # the headers of the signature are emtpy
        assert isinstance(
            signed_message, dict
        ), "signed_message must be a dict if attach_cose_key is True"
        key = signed_message.get("key")
        signed_message = signed_message.get("signature")  # type: ignore

    else:
        key = ""  # key will be extracted later from the payload headers

    # Add back the "D2" header byte and decode
    assert isinstance(
        signed_message, str
    ), "signed_message must be a hex string at this point"
    decoded_message = CoseMessage.decode(bytes.fromhex("d2" + signed_message))

    # generate/extract the cose key
    if not attach_cose_key:
        # get the verification key from the headers
        verification_key = decoded_message.phdr[KID]

        # generate the COSE key
        cose_key = {
            KpKty: KtyOKP,
            OKPKpCurve: Ed25519,
            KpKeyOps: [SignOp, VerifyOp],
            OKPKpX: verification_key,  # public key
        }

        cose_key = CoseKey.from_dict(cose_key)

    else:
        # i,e key is sent separately
        assert isinstance(
            key, str
        ), "key must be a hex string if attach_cose_key is True"
        cose_key = CoseKey.decode(bytes.fromhex(key))
        verification_key = cose_key[OKPKpX]

    # attach the key to the decoded message
    decoded_message.key = cose_key

    if len(verification_key) > 32:
        vk = BIP32ED25519PublicKey(
            public_key=verification_key[:32], chain_code=verification_key[32:]
        )
        vk.verify(
            signature=decoded_message.signature, message=decoded_message._sig_structure
        )
        signature_verified = True
    else:
        signature_verified = decoded_message.verify_signature()

    message = decoded_message.payload.decode("utf-8")

    signing_address = Address.from_primitive(decoded_message.phdr["address"])

    # check that the address attached matches the
    # one of the verification keys used to sign the message

    if signing_address.payment_part is not None:
        addresses_match = (
            signing_address.payment_part
            == PaymentVerificationKey.from_primitive(verification_key).hash()
        )
    else:
        addresses_match = (
            signing_address.staking_part
            == StakeVerificationKey.from_primitive(verification_key).hash()
        )

    verified = signature_verified & addresses_match

    return {
        "verified": verified,
        "message": message,
        "signing_address": signing_address,
    }



"""Transaction witness."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, List, Optional, Type, Union

from pprintpp import pformat

from pycardano.key import ExtendedVerificationKey, VerificationKey
from pycardano.nativescript import NativeScript
from pycardano.plutus import PlutusV1Script, PlutusV2Script, PlutusV3Script, Redeemers
from pycardano.serialization import (
    ArrayCBORSerializable,
    IndefiniteList,
    MapCBORSerializable,
    NonEmptyOrderedSet,
    limit_primitive_type,
)

__all__ = ["VerificationKeyWitness", "TransactionWitnessSet"]


@dataclass(repr=False)
class VerificationKeyWitness(ArrayCBORSerializable):
    vkey: Union[VerificationKey, ExtendedVerificationKey]
    signature: bytes

    @property
    def json_type(self) -> str:
        return "TxWitness ConwayEra"

    @property
    def json_description(self) -> str:
        return "Key Witness ShelleyEra"

    def __post_init__(self):
        # When vkey is in extended format, we need to convert it to non-extended, so it can match the
        # key hash of the input address we are trying to spend.
        if isinstance(self.vkey, ExtendedVerificationKey):
            self.vkey = self.vkey.to_non_extended()

    @classmethod
    @limit_primitive_type(list, tuple)
    def from_primitive(
        cls: Type[VerificationKeyWitness], values: Union[list, tuple]
    ) -> VerificationKeyWitness:
        return cls(
            vkey=VerificationKey.from_primitive(values[0]),
            signature=values[1],
        )

    def to_shallow_primitive(self) -> Union[list, tuple]:
        """Convert to a shallow primitive representation."""
        return [self.vkey.to_primitive(), self.signature]

    def __eq__(self, other):
        if not isinstance(other, VerificationKeyWitness):
            return False
        else:
            return (
                self.vkey.payload == other.vkey.payload
                and self.signature == other.signature
            )

    def __repr__(self):
        fields = {
            "vkey": self.vkey.payload.hex(),
            "signature": self.signature.hex(),
        }
        return pformat(fields, indent=2)


@dataclass(repr=False)
class TransactionWitnessSet(MapCBORSerializable):
    vkey_witnesses: Optional[
        Union[List[VerificationKeyWitness], NonEmptyOrderedSet[VerificationKeyWitness]]
    ] = field(
        default=None,
        metadata={
            "key": 0,
            "optional": True,
        },
    )

    native_scripts: Optional[
        Union[List[NativeScript], NonEmptyOrderedSet[NativeScript]]
    ] = field(
        default=None,
        metadata={
            "key": 1,
            "optional": True,
        },
    )

    # TODO: Add bootstrap witness (byron) support
    bootstrap_witness: Optional[List[Any]] = field(
        default=None, metadata={"optional": True, "key": 2}
    )

    plutus_v1_script: Optional[
        Union[List[PlutusV1Script], NonEmptyOrderedSet[PlutusV1Script]]
    ] = field(
        default=None,
        metadata={
            "key": 3,
            "optional": True,
        },
    )

    plutus_data: Optional[Union[List[Any], IndefiniteList, NonEmptyOrderedSet[Any]]] = (
        field(
            default=None,
            metadata={"optional": True, "key": 4},
        )
    )

    redeemer: Optional[Redeemers] = field(
        default=None,
        metadata={"optional": True, "key": 5},
    )

    plutus_v2_script: Optional[
        Union[List[PlutusV2Script], NonEmptyOrderedSet[PlutusV2Script]]
    ] = field(
        default=None,
        metadata={
            "key": 6,
            "optional": True,
        },
    )

    plutus_v3_script: Optional[
        Union[List[PlutusV3Script], NonEmptyOrderedSet[PlutusV3Script]]
    ] = field(
        default=None,
        metadata={
            "key": 7,
            "optional": True,
        },
    )

    def convert_to_latest_spec(self):
        # Convert lists to NonEmptyOrderedSet for fields that should use NonEmptyOrderedSet
        if isinstance(self.vkey_witnesses, list):
            self.vkey_witnesses = NonEmptyOrderedSet(self.vkey_witnesses)
        if isinstance(self.native_scripts, list):
            self.native_scripts = NonEmptyOrderedSet(self.native_scripts)
        if isinstance(self.plutus_data, list) and not isinstance(
            self.plutus_data, NonEmptyOrderedSet
        ):
            self.plutus_data = NonEmptyOrderedSet(list(self.plutus_data))
        if isinstance(self.plutus_v1_script, list):
            self.plutus_v1_script = NonEmptyOrderedSet(self.plutus_v1_script)
        if isinstance(self.plutus_v2_script, list):
            self.plutus_v2_script = NonEmptyOrderedSet(self.plutus_v2_script)
        if isinstance(self.plutus_v3_script, list):
            self.plutus_v3_script = NonEmptyOrderedSet(self.plutus_v3_script)

    def is_empty(self) -> bool:
        """Check if the witness set is empty."""
        return (
            not self.vkey_witnesses
            and not self.native_scripts
            and not self.bootstrap_witness
            and not self.plutus_v1_script
            and not self.plutus_data
            and not self.redeemer
            and not self.plutus_v2_script
            and not self.plutus_v3_script
        )
